# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/9/24 15:19
describe: 策略持仓权重管理
"""
import os
import time
import json
import threading
import pandas as pd
from loguru import logger
from datetime import datetime


class RedisWeightsClient:
    """策略持仓权重收发客户端"""

    version = "V240303"

    def __init__(self, strategy_name, redis_url=None, connection_pool=None, send_heartbeat=True, **kwargs):
        """
        :param strategy_name: str, 策略名
        :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取

            For example::

                redis://[[username]:[password]]@localhost:6379/0
                rediss://[[username]:[password]]@localhost:6379/0
                unix://[username@]/path/to/socket.sock?db=0[&password=password]

            Three URL schemes are supported:

            - `redis://` creates a TCP socket connection. See more at:
            <https://www.iana.org/assignments/uri-schemes/prov/redis>
            - `rediss://` creates a SSL wrapped TCP socket connection. See more at:
            <https://www.iana.org/assignments/uri-schemes/prov/rediss>
            - ``unix://``: creates a Unix Domain Socket connection.

        :param connection_pool: redis.BlockingConnectionPool, redis连接池，默认为None

            如果传入了 redis_url，则会自动创建一个连接池，否则需要传入一个连接池；如果传入了连接池，则会忽略 redis_url。

        :param send_heartbeat: boolean, 是否发送心跳，默认为True

            如果为True，会在后台启动一个线程，每15秒向redis发送一次心跳，用于检测策略是否存活。
            推荐在写入数据时设置为True，读取数据时设置为False，避免无用的心跳。

        :param kwargs: dict, 其他参数

            - key_prefix: str, redis中key的前缀，默认为 Weights
            - heartbeat_prefix: str, 心跳key的前缀，默认为 heartbeat
        """
        self.strategy_name = strategy_name
        self.key_prefix = kwargs.get("key_prefix", "Weights")

        if connection_pool:
            thread_safe_pool = connection_pool
            self.redis_url = connection_pool.connection_kwargs.get("url")
            logger.info(f"{strategy_name} {self.key_prefix}: 使用传入的 redis 连接池")

        else:
            import redis
            self.redis_url = redis_url if redis_url else os.getenv("RWC_REDIS_URL")
            thread_safe_pool = redis.BlockingConnectionPool.from_url(self.redis_url, decode_responses=True)
            logger.info(f"{strategy_name} {self.key_prefix}: 使用环境变量 RWC_REDIS_URL 创建 redis 连接池")

        import redis
        assert isinstance(thread_safe_pool, redis.BlockingConnectionPool), "redis连接池创建失败"

        self.r = redis.Redis(connection_pool=thread_safe_pool)
        self.lua_publish = RedisWeightsClient.register_lua_publish(self.r)
        self.heartbeat_prefix = kwargs.get("heartbeat_prefix", "heartbeat")

        if send_heartbeat:
            self.heartbeat_client = redis.Redis(connection_pool=thread_safe_pool)
            self.heartbeat_thread = threading.Thread(target=self.__heartbeat, daemon=True)
            self.heartbeat_thread.start()

    def set_metadata(self, base_freq, description, author, outsample_sdt, **kwargs):
        """设置策略元数据"""
        key = f"{self.key_prefix}:META:{self.strategy_name}"
        overwrite = kwargs.pop("overwrite", False)
        if self.r.exists(key):
            if not overwrite:
                logger.warning(f"已存在 {self.strategy_name} 的元数据，如需覆盖请设置 overwrite=True")
                return
            else:
                self.r.delete(key)
                logger.warning(f"删除 {self.strategy_name} 的元数据，重新写入")

        outsample_sdt = pd.to_datetime(outsample_sdt).strftime("%Y%m%d")
        meta = {
            "name": self.strategy_name,
            "base_freq": base_freq,
            "key_prefix": self.key_prefix,
            "description": description,
            "author": author,
            "outsample_sdt": outsample_sdt,
            "update_time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "kwargs": json.dumps(kwargs),
        }
        self.r.hset(key, mapping=meta)
        self.r.sadd(f"{self.key_prefix}:StrategyNames", self.strategy_name)

    def update_last(self, **kwargs):
        """设置策略最近一次更新时间，以及更新参数【可选】"""
        key = f"{self.key_prefix}:LAST:{self.strategy_name}"
        last = {
            "name": self.strategy_name,
            "time": datetime.now().strftime("%Y-%m-%d %H:%M:%S"),
            "kwargs": json.dumps(kwargs),
        }
        self.r.hset(key, mapping=last)

        logger.info(f"更新 {key} 的 last 时间")

    @property
    def metadata(self):
        """获取策略元数据"""
        key = f"{self.key_prefix}:META:{self.strategy_name}"
        return self.r.hgetall(key)

    @property
    def heartbeat_time(self):
        """获取策略的最近一次心跳时间"""
        key = f"{self.key_prefix}:{self.heartbeat_prefix}:{self.strategy_name}"
        return pd.to_datetime(self.r.get(key))

    def get_last_times(self, symbols=None):
        """获取所有品种上策略最近一次发布信号的时间

        :param symbols: list, 品种列表, 默认为None, 即获取所有品种
        :return: dict, {symbol: datetime}，如{'SFIF9001': datetime(2021, 9, 24, 15, 19, 0)}
        """
        if isinstance(symbols, str):
            row = self.r.hgetall(f"{self.key_prefix}:{self.strategy_name}:{symbols}:LAST")
            return pd.to_datetime(row["dt"]) if row else None  # type: ignore

        symbols = symbols if symbols else self.get_symbols()
        with self.r.pipeline() as pipe:
            for symbol in symbols:
                pipe.hgetall(f"{self.key_prefix}:{self.strategy_name}:{symbol}:LAST")
            rows = pipe.execute()
        return {x["symbol"]: pd.to_datetime(x["dt"]) for x in rows}

    def publish(self, symbol, dt, weight, price=0, ref=None, overwrite=False):
        """发布单个策略持仓权重

        :param symbol: str, eg; SFIF9001
        :param dt: py_datetime or pandas Timestamp
        :param weight: float, 信号值
        :param price: float, 产生信号时的价格
        :param ref: dict, 自定义数据
        :param overwrite: boolean, 是否覆盖已有记录
        :return: 成功发布信号的条数
        """
        if not isinstance(dt, datetime):
            dt = pd.to_datetime(dt)

        if not overwrite:
            last_dt = self.get_last_times(symbol)
            if last_dt is not None and dt <= last_dt:  # type: ignore
                logger.warning(f"不允许重复写入，已过滤 {symbol} {dt} 的重复信号")
                return 0

        udt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        key = f'{self.key_prefix}:{self.strategy_name}:{symbol}:{dt.strftime("%Y%m%d%H%M%S")}'
        ref = ref if ref else "{}"
        ref_str = json.dumps(ref) if isinstance(ref, dict) else ref
        return self.lua_publish(keys=[key], args=[1 if overwrite else 0, udt, weight, price, ref_str])

    def publish_dataframe(self, df, overwrite=False, batch_size=10000):
        """批量发布多个策略信号

        :param df: pandas.DataFrame, 必需包含['symbol', 'dt', 'weight']列,
                可选['price', 'ref']列, 如没有price则写0, dtype同publish方法
        :param overwrite: boolean, 是否覆盖已有记录
        :param batch_size: int, 每次发布的最大数量
        :return: 成功发布信号的条数
        """
        df = df.copy()
        df["dt"] = pd.to_datetime(df["dt"])
        logger.info(f"输入数据中有 {len(df)} 条权重信号")

        # 去除单个品种下相邻时间权重相同的数据
        _res = []
        for _, dfg in df.groupby("symbol"):
            dfg = dfg.sort_values("dt", ascending=True).reset_index(drop=True)
            dfg = dfg[dfg["weight"].diff().fillna(1) != 0].copy()
            _res.append(dfg)
        df = pd.concat(_res, ignore_index=True)
        df = df.sort_values(["dt"]).reset_index(drop=True)
        logger.info(f"去除单个品种下相邻时间权重相同的数据后，剩余 {len(df)} 条权重信号")

        if "price" not in df.columns:
            df["price"] = 0
        if "ref" not in df.columns:
            df["ref"] = "{}"

        if not overwrite:
            raw_count = len(df)
            _time = self.get_last_times()
            _data = []
            for symbol, dfg in df.groupby("symbol"):
                last_dt = _time.get(symbol)
                if last_dt is not None:
                    dfg = dfg[dfg["dt"] > last_dt]
                _data.append(dfg)
            df = pd.concat(_data, ignore_index=True)
            logger.info(f"不允许重复写入，已过滤 {raw_count - len(df)} 条重复信号")

        keys, args = [], []
        for row in df[["symbol", "dt", "weight", "price", "ref"]].to_numpy():
            key = f'{self.key_prefix}:{self.strategy_name}:{row[0]}:{row[1].strftime("%Y%m%d%H%M%S")}'
            keys.append(key)

            args.append(row[2])
            args.append(row[3])
            ref = row[4]
            args.append(json.dumps(ref) if isinstance(ref, dict) else ref)

        udt = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        overwrite = 1 if overwrite else 0

        pub_cnt = 0
        len_keys = len(keys)
        for i in range(0, len_keys, batch_size):
            if i + batch_size < len_keys:
                tmp_keys = keys[i : i + batch_size]
                tmp_args = [overwrite, udt] + args[3 * i : 3 * (i + batch_size)]
            else:
                tmp_keys = keys[i:len_keys]
                tmp_args = [overwrite, udt] + args[3 * i : 3 * len_keys]
            logger.info(f"索引 {i}，即将发布 {len(tmp_keys)} 条权重信号")
            pub_cnt += self.lua_publish(keys=tmp_keys, args=tmp_args)
            logger.info(f"已完成 {pub_cnt} 次发布")

        self.update_last()
        return pub_cnt

    def __heartbeat(self):
        while True:
            key = f"{self.key_prefix}:{self.heartbeat_prefix}:{self.strategy_name}"
            try:
                self.heartbeat_client.set(key, datetime.now().strftime("%Y-%m-%d %H:%M:%S"))
            except Exception as e:
                logger.error(f"心跳发送失败：{e}")
            time.sleep(15)

    def get_keys(self, pattern) -> list:
        """获取 redis 中指定 pattern 的 keys"""
        k = self.r.keys(pattern)
        return k if k else []  # type: ignore

    def clear_all(self, with_human=True):
        """删除该策略所有记录"""
        keys = self.get_keys(f"{self.key_prefix}:{self.strategy_name}*")
        keys.append(f"{self.key_prefix}:META:{self.strategy_name}")
        keys.append(f"{self.key_prefix}:LAST:{self.strategy_name}")
        keys.append(f"{self.key_prefix}:{self.heartbeat_prefix}:{self.strategy_name}")

        if len(keys) == 0:
            logger.warning(f"{self.strategy_name} 没有记录")
            return

        if with_human:
            human = input(f"{self.strategy_name} 即将删除 {len(keys)} 条记录，是否确认？(y/n):")
            if human.lower() != "y":
                logger.warning(f"{self.strategy_name} 删除操作已取消")
                return

        self.r.delete(*keys)  # type: ignore
        cong = self.r.srem(f"{self.key_prefix}:StrategyNames", self.strategy_name)
        logger.info(f"{self.strategy_name} 删除了 {len(keys)} 条记录, {cong} 条策略名")

    @staticmethod
    def register_lua_publish(client):
        lua_body = """
local overwrite = ARGV[1]
local update_time = ARGV[2]
local cnt = 0
local ret
for i = 1, #KEYS do
    local key = KEYS[i]
    local sig = ARGV[3 + 3 * (i - 1)]
    local price = ARGV[4 + 3 * (i - 1)]
    local ref_str = ARGV[5 + 3 * (i - 1)]
    local split_str = {}
    key:gsub('[^:]+', function(s) table.insert(split_str, s) end)
    local model_key = split_str[1] .. ':' .. split_str[2] .. ':' .. split_str[3]

    local if_pass = true
    if overwrite ~= '0' then
        if_pass = false
    else
        local pos = redis.call('HGET', model_key .. ':LAST', 'weight')
        if not pos or math.abs(tonumber(sig) - tonumber(pos)) > 0.00001 then
            if_pass = false
        end
    end

    if not if_pass then
        local strategy_name, symbol, action_time = split_str[2], split_str[3], split_str[4]
        local at_str = string.sub(action_time, 1, 4) .. '-' .. string.sub(action_time, 5, 6) .. '-' ..
            string.sub(action_time, 7, 8) .. ' ' .. string.sub(action_time, 9, 10) .. ':' ..
            string.sub(action_time, 11, 12) .. ':' .. string.sub(action_time, 13, 14)
        redis.call('ZADD', model_key, tonumber(action_time), key)
        local ret1 = redis.call('HMSET', key, 'symbol', symbol, 'weight', sig, 'dt', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        local ret2 = redis.call('HMSET', key:gsub(action_time, 'LAST'), 'symbol', symbol, 'weight', sig, 'dt', at_str, 'update_time', update_time, 'price', price, 'ref', ref_str)
        if ret1.ok and ret2.ok then
            cnt = cnt + 1
            local pubKey = 'PUBSUB:' .. split_str[1] .. ':' .. strategy_name .. ':' .. symbol
            redis.call('PUBLISH', pubKey, key .. ':' .. sig .. ':' .. price .. ':' .. ref_str)
        end
    end
end
return cnt
"""
        return client.register_script(lua_body)

    def get_symbols(self):
        """获取策略交易的品种列表"""
        keys = self.get_keys(f"{self.key_prefix}:{self.strategy_name}:*:LAST")
        symbols = {x.split(":")[2] for x in keys}  # type: ignore
        return list(symbols)

    def get_last_weights(self, symbols=None, ignore_zero=True, lua=True):
        """获取最近的持仓权重

        :param symbols: list, 品种列表
        :param ignore_zero: boolean, 是否忽略权重为0的品种
        :param lua: boolean, 是否使用 lua 脚本获取，默认为True
            如果要全量获取，推荐使用 lua 脚本，速度更快；如果要获取指定 symbols，不推荐使用 lua 脚本。
        :return: pd.DataFrame
        """
        if lua:
            lua_script = """
            local keys = redis.call('KEYS', ARGV[1])
            local results = {}
            for i=1, #keys do
                results[i] = redis.call('HGETALL', keys[i])
            end
            return results
            """
            key_pattern = self.key_prefix + ":" + self.strategy_name + ":*:LAST"
            results = self.r.eval(lua_script, 0, key_pattern)
            rows = [dict(zip(r[::2], r[1::2])) for r in results]  # type: ignore
            if symbols:
                rows = [r for r in rows if r["symbol"] in symbols]

        else:
            symbols = symbols if symbols else self.get_symbols()
            with self.r.pipeline() as pipe:
                for symbol in symbols:
                    pipe.hgetall(f"{self.key_prefix}:{self.strategy_name}:{symbol}:LAST")
                rows = pipe.execute()

        dfw = pd.DataFrame(rows)
        dfw["weight"] = dfw["weight"].astype(float)
        dfw["dt"] = pd.to_datetime(dfw["dt"])
        if ignore_zero:
            dfw = dfw[dfw["weight"] != 0].copy().reset_index(drop=True)
        dfw = dfw.sort_values(["dt", "symbol"]).reset_index(drop=True)
        return dfw

    def get_hist_weights(self, symbol, sdt, edt) -> pd.DataFrame:
        """获取单个品种的持仓权重历史数据

        :param symbol: str, 品种代码
        :param sdt: str, 开始时间, eg: 20210924 10:19:00
        :param edt: str, 结束时间, eg: 20220924 10:19:00
        :return: pd.DataFrame
        """
        start_score = pd.to_datetime(sdt).strftime("%Y%m%d%H%M%S")
        end_score = pd.to_datetime(edt).strftime("%Y%m%d%H%M%S")
        model_key = f"{self.key_prefix}:{self.strategy_name}:{symbol}"
        key_list = self.r.zrangebyscore(model_key, start_score, end_score)

        if len(key_list) == 0:
            logger.warning(f"no history weights: {symbol} - {sdt} - {edt}")
            return pd.DataFrame()

        with self.r.pipeline() as pipe:
            for key in key_list:
                pipe.hmget(key, "weight", "price", "ref")
            rows = pipe.execute()

        weights = []
        for i in range(len(key_list)):
            dt = pd.to_datetime(key_list[i].split(":")[-1])
            weight, price, ref = rows[i]
            weight = weight if weight is None else float(weight)
            price = price if price is None else float(price)
            try:
                ref = json.loads(ref)
            except Exception as e:
                ref = ref
            weights.append((self.strategy_name, symbol, dt, weight, price, ref))

        dfw = pd.DataFrame(weights, columns=["strategy_name", "symbol", "dt", "weight", "price", "ref"])
        dfw = dfw.sort_values("dt").reset_index(drop=True)
        return dfw

    def get_all_weights(self, sdt=None, edt=None, **kwargs) -> pd.DataFrame:
        """获取所有权重数据

        :param sdt: str, 开始时间, eg: 20210924 10:19:00
        :param edt: str, 结束时间, eg: 20220924 10:19:00
        :return: pd.DataFrame
        """
        lua_script = """
        local keys = redis.call('KEYS', ARGV[1])
        local results = {}
        for i=1, #keys do
            local last_part = keys[i]:match('([^:]+)$')
            if #last_part == 14 and tonumber(last_part) ~= nil then
                results[#results + 1] = redis.call('HGETALL', keys[i])
            end
        end
        return results
        """
        key_pattern = self.key_prefix + ":" + self.strategy_name + ":*:*"
        results = self.r.eval(lua_script, 0, key_pattern)
        results = [dict(zip(r[::2], r[1::2])) for r in results]  # type: ignore

        df = pd.DataFrame(results)
        df["dt"] = pd.to_datetime(df["dt"])
        df["weight"] = df["weight"].astype(float)
        df = df.sort_values(["dt", "symbol"]).reset_index(drop=True)
        # df 中的columns：['symbol', 'weight', 'dt', 'update_time', 'price', 'ref']

        df1 = pd.pivot_table(df, index="dt", columns="symbol", values="weight").sort_index().ffill().fillna(0)
        df1 = pd.melt(df1.reset_index(), id_vars="dt", value_vars=df1.columns, value_name="weight")  # type: ignore

        # 加上 df 中的 update_time 信息
        df1 = df1.merge(df[["dt", "symbol", "update_time"]], on=["dt", "symbol"], how="left")
        df1 = df1.sort_values(["symbol", "dt"]).reset_index(drop=True)
        for _, dfg in df1.groupby("symbol"):
            df1.loc[dfg.index, "update_time"] = dfg["update_time"].ffill().bfill()

        if sdt:
            df1 = df1[df1["dt"] >= pd.to_datetime(sdt)].reset_index(drop=True)
        if edt:
            df1 = df1[df1["dt"] <= pd.to_datetime(edt)].reset_index(drop=True)
        df1 = df1.sort_values(["dt", "symbol"]).reset_index(drop=True)
        return df1


def clear_strategy(strategy_name, redis_url=None, connection_pool=None, key_prefix="Weights", **kwargs):
    """删除策略所有记录

    :param strategy_name: str, 策略名
    :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取
    :param connection_pool: redis.ConnectionPool, redis连接池
    :param key_prefix: str, redis中key的前缀，默认为 Weights
    :param kwargs: dict, 其他参数
    """
    with_human = kwargs.pop("with_human", True)
    rwc = RedisWeightsClient(
        strategy_name,
        redis_url=redis_url,
        connection_pool=connection_pool,
        key_prefix=key_prefix,
        send_heartbeat=False,
        **kwargs,
    )
    rwc.clear_all(with_human)


def get_strategy_weights(strategy_name, redis_url=None, connection_pool=None, key_prefix="Weights", **kwargs):
    """获取策略的持仓权重

    :param strategy_name: str, 策略名
    :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取
    :param connection_pool: redis.ConnectionPool, redis连接池
    :param key_prefix: str, redis中key的前缀，默认为 Weights
    :param kwargs: dict, 其他参数

        - symbols : list, 品种列表, 默认为None, 即获取所有品种的权重
        - sdt : str, 开始时间, eg: 20210924 10:19:00
        - edt : str, 结束时间, eg: 20220924 10:19:00
        - only_last : boolean, 是否只保留每个品种最近一次权重，默认为False

    :return: pd.DataFrame
    """
    kwargs.pop("send_heartbeat", None)
    rwc = RedisWeightsClient(
        strategy_name,
        redis_url=redis_url,
        connection_pool=connection_pool,
        key_prefix=key_prefix,
        send_heartbeat=False,
        **kwargs,
    )
    sdt = kwargs.get("sdt")
    edt = kwargs.get("edt")
    symbols = kwargs.get("symbols")
    only_last = kwargs.get("only_last", False)

    if only_last:
        # 保留每个品种最近一次权重
        df = rwc.get_last_weights(ignore_zero=False)
        return df

    df = rwc.get_all_weights(sdt=sdt, edt=edt)
    if symbols:
        # 保留指定品种的权重
        not_in = [x for x in symbols if x not in df["symbol"].unique()]
        if not_in:
            logger.warning(f"{strategy_name} 中没有 {not_in} 的权重记录")

        df = df[df["symbol"].isin(symbols)].reset_index(drop=True)

    return df


def get_strategy_mates(redis_url=None, connection_pool=None, key_pattern="Weights:META:*", **kwargs):
    """获取Redis中的策略元数据

    :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取
    :param connection_pool: redis.ConnectionPool, redis连接池
    :param key_pattern: str, redis中key的pattern，默认为 Weights:META:*
    :param kwargs: dict, 其他参数
    :return: pd.DataFrame
    """
    heartbeat_prefix = kwargs.get("heartbeat_prefix", "heartbeat")

    import redis

    if connection_pool:
        r = redis.Redis(connection_pool=connection_pool)
    else:
        redis_url = redis_url if redis_url else os.getenv("RWC_REDIS_URL")
        r = redis.Redis.from_url(redis_url, decode_responses=True)

    rows = []
    for key in r.keys(key_pattern):  # type: ignore
        meta = r.hgetall(key)
        if not meta:
            logger.warning(f"{key} 没有策略元数据")
            continue

        meta["heartbeat_time"] = r.get(f"{meta['key_prefix']}:{heartbeat_prefix}:{meta['name']}")  # type: ignore
        rows.append(meta)

    if len(rows) == 0:
        logger.warning(f"{key_pattern} 下没有策略元数据")
        return pd.DataFrame()

    df = pd.DataFrame(rows)
    df["update_time"] = pd.to_datetime(df["update_time"])
    df["heartbeat_time"] = pd.to_datetime(df["heartbeat_time"])
    df = df.sort_values("name").reset_index(drop=True)

    r.close()
    return df


def get_heartbeat_time(strategy_name=None, redis_url=None, connection_pool=None, key_prefix="Weights", **kwargs):
    """获取策略的最近一次心跳时间

    :param strategy_name: str, 策略名，默认为None, 即获取所有策略的心跳时间
    :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取
    :param connection_pool: redis.ConnectionPool, redis连接池
    :param key_prefix: str, redis中key的前缀，默认为 Weights
    :param kwargs: dict, 其他参数

            - heartbeat_prefix: str, 心跳key的前缀，默认为 heartbeat

    :return: str, 最近一次心跳时间
    """
    import redis

    if connection_pool:
        r = redis.Redis(connection_pool=connection_pool)
    else:
        redis_url = redis_url if redis_url else os.getenv("RWC_REDIS_URL")
        r = redis.Redis.from_url(redis_url, decode_responses=True)

    if not strategy_name:
        dfm = get_strategy_mates(
            redis_url=redis_url, connection_pool=connection_pool, key_pattern=f"{key_prefix}:META:*"
        )
        if len(dfm) == 0:
            logger.warning(f"{key_prefix} 下没有策略元数据")
            return None
        strategy_names = dfm["name"].unique().tolist()
    else:
        strategy_names = [strategy_name]

    heartbeat_prefix = kwargs.get("heartbeat_prefix", "heartbeat")
    res = {}
    for sn in strategy_names:
        hdt = r.get(f"{key_prefix}:{heartbeat_prefix}:{sn}")
        if hdt:
            res[sn] = pd.to_datetime(hdt)
        else:
            res[sn] = None
            logger.warning(f"{sn} 没有心跳时间")
    r.close()
    return res


def get_strategy_names(redis_url=None, connection_pool=None, key_prefix="Weights"):
    """获取所有策略名的列表.

    :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取
    :param connection_pool: redis.ConnectionPool, redis连接池
    :param key_prefix: str, redis中key的前缀，默认为 Weights
    :return: list, 所有策略名
    """
    import redis

    if connection_pool:
        r = redis.Redis(connection_pool=connection_pool)
    else:
        redis_url = redis_url if redis_url else os.getenv("RWC_REDIS_URL")
        r = redis.Redis.from_url(redis_url, decode_responses=True)

    rs = r.smembers(f"{key_prefix}:StrategyNames")
    return list(rs)


def get_strategy_latest(redis_url=None, connection_pool=None, key_prefix="Weights"):
    """获取所有策略的最新持仓

    :param redis_url: str, redis连接字符串, 默认为None, 即从环境变量 RWC_REDIS_URL 中读取
    :param connection_pool: redis.ConnectionPool, redis连接池
    :param key_prefix: str, redis中key的前缀，默认为 Weights
    :return: pd.DataFrame, 最新持仓数据
    """
    import redis

    if connection_pool:
        r = redis.Redis(connection_pool=connection_pool)
    else:
        redis_url = redis_url if redis_url else os.getenv("RWC_REDIS_URL")
        r = redis.Redis.from_url(redis_url, decode_responses=True)

    keys = r.keys(f"{key_prefix}:*:LAST")
    pipeline = r.pipeline()
    for key in keys:
        pipeline.hgetall(key)
    fetched_data = pipeline.execute()

    results = []
    for key, data in zip(keys, fetched_data):
        strategy = key.split(":")[1]
        data["strategy"] = strategy
        results.append(data)

    df = pd.DataFrame(results)
    df["weight"] = df["weight"].astype(float)
    df = df[["dt", "strategy", "symbol", "weight", "update_time"]]
    return df



import pandas as pd

def risk_free_returns(start_date: str = '20180101', end_date: str = '20210101', year_returns: float = 0.03): ...
def update_nxb(df: pd.DataFrame, **kwargs) -> pd.DataFrame: ...
def update_bbars(da, price_col: str = 'close', numbers=(1, 2, 5, 10, 20, 30)) -> None: ...
def update_tbars(da: pd.DataFrame, event_col: str) -> None: ...
def resample_to_daily(df: pd.DataFrame, sdt=None, edt=None, only_trade_date: bool = True): ...
def adjust_holding_weights(df, hold_periods: int = 1, **kwargs): ...



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2019/10/29 15:01
"""
from rs_czsc import (
    daily_performance,
    top_drawdowns,
    WeightBacktest,
)

from czsc import envs
from czsc import utils
from czsc import traders
from czsc.core import CZSC, Freq, Operate, Direction, Signal, Event, RawBar, NewBar, Position, ZS, format_standard_kline
from czsc.utils import ta
from czsc.traders import (
    CzscTrader,
    CzscSignals,
    generate_czsc_signals,
    check_signals_acc,
    get_unique_signals,
    PairsPerformance,
    combine_holds_and_pairs,
    combine_dates_and_pairs,
    DummyBacktest,
    SignalsParser,
    get_signals_config,
    get_signals_freqs,
    stoploss_by_direction,
    get_ensemble_weight,
    OpensOptimize,
    ExitsOptimize,
)

from czsc.utils import (
    timeout_decorator,
    mac_address,
    overlap,
    to_arrow,
    dill_dump,
    dill_load,
    read_json,
    save_json,
    get_sub_elements,
    get_py_namespace,
    code_namespace,
    freqs_sorted,
    x_round,
    import_by_name,
    create_grid_params,
    update_bbars,
    update_tbars,
    update_nxb,
    risk_free_returns,
    resample_to_daily,
    cross_sectional_ic,
    rolling_daily_performance,
    holds_performance,
    psi,
    home_path,
    DiskCache,
    disk_cache,
    clear_cache,
    clear_expired_cache,
    get_dir_size,
    empty_cache_path,
    print_df_sample,
    index_composition,
    AliyunOSS,
    DataClient,
    set_url_token,
    get_url_token,
    generate_fernet_key,
    fernet_encrypt,
    fernet_decrypt,
)


__version__ = "0.10.12"
__author__ = "zengbin93"
__email__ = "zeng_bin8888@163.com"
__date__ = "20260308"

# 延迟加载重型可选模块，避免影响导入速度
_LAZY_MODULES = {
    "svc": "czsc.svc",
    "fsa": "czsc.fsa",
    "sensors": "czsc.sensors",
    "aphorism": "czsc.aphorism",
    "mock": "czsc.mock",
    "rwc": "czsc.traders.rwc",
    "cwc": "czsc.traders.cwc",
}

# 延迟加载的属性映射：属性名 -> (模块路径, 属性名)
_LAZY_ATTRS = {
    # czsc.strategies
    'CzscStrategyBase': ('czsc.strategies', 'CzscStrategyBase'),
    'CzscJsonStrategy': ('czsc.strategies', 'CzscJsonStrategy'),
    # czsc.sensors
    'holds_concepts_effect': ('czsc.sensors', 'holds_concepts_effect'),
    'CTAResearch': ('czsc.sensors', 'CTAResearch'),
    # czsc.utils.warning_capture
    'capture_warnings': ('czsc.utils.warning_capture', 'capture_warnings'),
    'execute_with_warning_capture': ('czsc.utils.warning_capture', 'execute_with_warning_capture'),
    # czsc.py.calendar
    'is_trading_date': ('czsc.py.calendar', 'is_trading_date'),
    'next_trading_date': ('czsc.py.calendar', 'next_trading_date'),
    'prev_trading_date': ('czsc.py.calendar', 'prev_trading_date'),
    'get_trading_dates': ('czsc.py.calendar', 'get_trading_dates'),
    # czsc.utils.trade
    'adjust_holding_weights': ('czsc.utils.trade', 'adjust_holding_weights'),
    # czsc.utils.log
    'log_strategy_info': ('czsc.utils.log', 'log_strategy_info'),
    # czsc.utils.bi_info
    'calculate_bi_info': ('czsc.utils.bi_info', 'calculate_bi_info'),
    'symbols_bi_infos': ('czsc.utils.bi_info', 'symbols_bi_infos'),
    # czsc.features.utils
    'is_event_feature': ('czsc.features.utils', 'is_event_feature'),
    'normalize_corr': ('czsc.features.utils', 'normalize_corr'),
    'feature_returns': ('czsc.features.utils', 'feature_returns'),
    'feature_sectional_corr': ('czsc.features.utils', 'feature_sectional_corr'),
    # czsc.utils.plotting.kline
    'plot_czsc_chart': ('czsc.utils.plotting.kline', 'plot_czsc_chart'),
    'KlineChart': ('czsc.utils.plotting.kline', 'KlineChart'),
    # czsc.utils.kline_quality
    'check_kline_quality': ('czsc.utils.kline_quality', 'check_kline_quality'),
    # czsc.py.bar_generator
    'resample_bars': ('czsc.py.bar_generator', 'resample_bars'),
    'get_intraday_times': ('czsc.py.bar_generator', 'get_intraday_times'),
    'check_freq_and_market': ('czsc.py.bar_generator', 'check_freq_and_market'),
    # czsc.eda
    'remove_beta_effects': ('czsc.eda', 'remove_beta_effects'),
    'vwap': ('czsc.eda', 'vwap'),
    'twap': ('czsc.eda', 'twap'),
    'cross_sectional_strategy': ('czsc.eda', 'cross_sectional_strategy'),
    'judge_factor_direction': ('czsc.eda', 'judge_factor_direction'),
    'monotonicity': ('czsc.eda', 'monotonicity'),
    'min_max_limit': ('czsc.eda', 'min_max_limit'),
    'rolling_layers': ('czsc.eda', 'rolling_layers'),
    'cal_symbols_factor': ('czsc.eda', 'cal_symbols_factor'),
    'weights_simple_ensemble': ('czsc.eda', 'weights_simple_ensemble'),
    'unify_weights': ('czsc.eda', 'unify_weights'),
    'sma_long_bear': ('czsc.eda', 'sma_long_bear'),
    'dif_long_bear': ('czsc.eda', 'dif_long_bear'),
    'tsf_type': ('czsc.eda', 'tsf_type'),
    'limit_leverage': ('czsc.eda', 'limit_leverage'),
    'cal_trade_price': ('czsc.eda', 'cal_trade_price'),
    'mark_cta_periods': ('czsc.eda', 'mark_cta_periods'),
    'mark_volatility': ('czsc.eda', 'mark_volatility'),
    'cal_yearly_days': ('czsc.eda', 'cal_yearly_days'),
    'turnover_rate': ('czsc.eda', 'turnover_rate'),
    'make_price_features': ('czsc.eda', 'make_price_features'),
    # czsc.utils.backtest_report
    'generate_backtest_report': ('czsc.utils.backtest_report', 'generate_backtest_report'),
}


def __getattr__(name):
    import importlib

    if name in _LAZY_MODULES:
        module = importlib.import_module(_LAZY_MODULES[name])
        globals()[name] = module
        return module

    if name in _LAZY_ATTRS:
        mod_path, attr_name = _LAZY_ATTRS[name]
        module = importlib.import_module(mod_path)
        attr = getattr(module, attr_name)
        globals()[name] = attr
        return attr

    raise AttributeError(f"module 'czsc' has no attribute {name!r}")


def welcome():
    from czsc import aphorism

    print(f"欢迎使用CZSC！当前版本标识为 {__version__}@{__date__}\n")
    aphorism.print_one()

    print(
        f"CZSC环境变量：" f"czsc_min_bi_len = {envs.get_min_bi_len()}; " f"czsc_max_bi_num = {envs.get_max_bi_num()}; "
    )
    if get_dir_size(home_path) > pow(1024, 3):
        print(f"{home_path} 目录缓存超过1GB，请适当清理。调用 czsc.empty_cache_path() 可以直接清空缓存")


# if get_dir_size(home_path) > pow(1024, 3):
#     print(f"{home_path} 目录缓存超过1GB，请适当清理。调用 czsc.empty_cache_path() 可以直接清空缓存")



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/9/24 13:33
describe: 择时策略开平仓优化
"""
import os
import czsc
import time
import hashlib
import pandas as pd
from tqdm import tqdm
from loguru import logger
from copy import deepcopy
from pathlib import Path
from typing import Callable, Union, List, AnyStr
from czsc.strategies import CzscStrategyBase
from czsc.core import Position, Event, Signal
from concurrent.futures import ProcessPoolExecutor, as_completed


class CzscOpenOptimStrategy(CzscStrategyBase):

    @staticmethod
    def update_beta_opens(beta: Position, open_signals_all: Union[List[AnyStr], AnyStr]):
        """更新 beta 出场信号"""
        pos = deepcopy(beta)
        assert len(pos.opens) == 1, "基础策略入场信号必须为单个Event"
        if isinstance(open_signals_all, str):
            open_signals_all = [open_signals_all]

        open_signals_all = [Signal(s) for s in open_signals_all]
        open_signals_all = [{"key": s.key, "value": s.value} for s in open_signals_all]
        pos_dict = pos.dump()
        sig_hash = hashlib.md5(f"{open_signals_all}".encode('utf-8')).hexdigest()[:8].upper()
        pos_dict['name'] = f"{pos.name}#{sig_hash}"
        pos_dict['opens'][0]['signals_all'].extend(open_signals_all)
        return Position.load(pos_dict)

    @property
    def positions(self):
        betas = self.load_positions(self.kwargs['files_position'])
        candidate_signals = deepcopy(self.kwargs['candidate_signals'])
        pos_list = deepcopy(betas)
        for beta in betas:
            for sigs_ in candidate_signals:
                pos = self.update_beta_opens(beta, sigs_)
                pos_list.append(pos)
        return pos_list


class CzscExitOptimStrategy(CzscStrategyBase):

    @staticmethod
    def update_beta_exits(beta: Position, event_dict: dict, mode='replace'):
        """更新 beta 出场信号

        出场优化，有三种路径：

        1. 开仓后识别走势加速，找个合适的机会提前平仓
        2. 根据开仓后一段时间内的走势，如果发现异样，提前平仓
        3. 用新的出场信号完全替换旧的出场信号

        """
        pos = deepcopy(beta)
        event: Event = Event.load(deepcopy(event_dict))
        event_hash = hashlib.md5(f"{event.dump()}".encode('utf-8')).hexdigest()[:8].upper()
        open_ops = [x.operate.value for x in pos.opens]

        if all(x == '开多' for x in open_ops) and event.operate.value != '平多':
            return None

        if all(x == '开空' for x in open_ops) and event.operate.value != '平空':
            return None

        assert mode in ['replace', 'append'], "mode must be replace or append"
        if mode == 'replace':
            pos.exits = [deepcopy(event)]
            pos.name = f"{beta.name}#替换{event_hash}"
        else:
            pos.exits.append(deepcopy(event))
            pos.name = f"{beta.name}#追加{event_hash}"

        # 踩坑记录：对于 dataclass 对象，如果直接修改属性，会导致原对象也被修改，因此这里需要重新创建一个对象
        return Position.load(pos.dump())

    @property
    def positions(self):
        betas = self.load_positions(self.kwargs['files_position'])
        candidate_events = deepcopy(self.kwargs['candidate_events'])  # 优化出场信号
        pos_list = deepcopy(betas)

        for beta in betas:
            for event in candidate_events:

                pos1 = self.update_beta_exits(beta, event, mode='append')
                pos2 = self.update_beta_exits(beta, event, mode='replace')

                if pos1 is not None:
                    pos_list.append(pos1)
                if pos2 is not None:
                    pos_list.append(pos2)

        return pos_list


def one_symbol_optim(symbol, read_bars: Callable, path: str, **kwargs):
    """单个标的优化

    :param symbol: 标的代码
    :param read_bars: K线数据读取函数
    :param path: 优化结果保存路径
    :param kwargs: 其他参数

        - bar_sdt: K线数据开始日期
        - bar_edt: K线数据结束日期
        - sdt: 优化开始日期
        - optim_type: 优化类型，open 或 exit

    """
    symbol_path = Path(path) / symbol
    if symbol_path.exists():
        logger.info(f"{symbol} dummy 结果已经存在")
        return
    symbol_path.mkdir(parents=True, exist_ok=True)

    bar_sdt = kwargs.get('bar_sdt', '20150101')
    bar_edt = kwargs.get('bar_edt', '20220101')
    sdt = kwargs.get('sdt', '20170101')
    assert bar_sdt < sdt < bar_edt, "sdt 必须在 bar_sdt 和 bar_edt 之间"

    start_time = time.time()
    optim_type = kwargs.get('optim_type', 'open')
    if optim_type == 'open':
        tactic = CzscOpenOptimStrategy(symbol=symbol, **kwargs)
    else:
        assert optim_type == 'exit', "optim_type must be open or exit"
        tactic = CzscExitOptimStrategy(symbol=symbol, **kwargs)

    try:
        bars = read_bars(symbol, tactic.base_freq, bar_sdt, bar_edt, fq='后复权', raw_bar=True)
        if len(bars) < 100:
            logger.warning(f"{symbol} K线数量不足，无法优化")
            return None

        trader = tactic.backtest(bars, sdt=sdt)
    except Exception as e:
        logger.exception(f"{symbol} 优化失败，原因：{e}")
        return None

    for pos in trader.positions:            # type: ignore
        try:
            file_pairs = os.path.join(symbol_path, f"{pos.name}.pairs.feather")
            file_holds = os.path.join(symbol_path, f"{pos.name}.holds.feather")

            pairs = pd.DataFrame(pos.pairs)
            pairs.to_feather(file_pairs)

            dfh = pd.DataFrame(pos.holds)
            if not dfh.empty:   
                dfh['n1b'] = (dfh['price'].shift(-1) / dfh['price'] - 1) * 10000
                dfh.fillna(0, inplace=True)
                dfh['symbol'] = pos.symbol
            dfh.to_feather(file_holds)
        except Exception as e:
            logger.debug(f"{symbol} {pos.name} 保存失败，原因：{e}")

    logger.info(f"{symbol} - {optim_type} 优化完成，耗时 {time.time() - start_time:.2f} 秒")


def one_position_stats(path, pos_name):
    """分析单个 pos 的表现"""
    path = Path(path)
    files_p = path.glob(f"**/{pos_name}.pairs.feather")
    files_h = path.glob(f"**/{pos_name}.holds.feather")

    pos_pairs = []
    for file_p in files_p:
        try:
            dfp = pd.read_feather(file_p)
            pos_pairs.append(dfp)
        except Exception as e:
            logger.debug(f"{file_p} 读取失败，原因：{e}")

    pos_holds = []
    for file_h in files_h:
        try:
            dfh = pd.read_feather(file_h)
            pos_holds.append(dfh)
        except Exception as e:
            logger.debug(f"{file_h} 读取失败，原因：{e}")

    pairs = pd.concat(pos_pairs, ignore_index=True)
    holds = pd.concat(pos_holds, ignore_index=True)

    if len(pairs) == 0 or len(holds) == 0:
        return None

    try:
        pp = czsc.PairsPerformance(pairs)
        stats = dict(pp.basic_info)
        # 加入截面等权评价
        cross = holds.groupby('dt', group_keys=False).apply(lambda x: (x['n1b'] * x['pos']).sum() / (sum(x['pos'] != 0) + 1), include_groups=False).sum()
        stats['截面等权收益'] = cross
        cross1 = holds.groupby('dt', group_keys=False).apply(lambda x: (x['n1b'] * x['pos']).mean(), include_groups=False).sum()
        stats['截面品种等权'] = cross1
        stats['pos_name'] = pos_name
        return stats
    except Exception as e:
        logger.exception(f"{pos_name} 分析失败，原因：{e}")
        return None


class OpensOptimize:
    """基础策略入场优化流程"""

    def __init__(self, read_bars: Callable, **kwargs):
        """

        :param read_bars: K线数据读取函数
        :param kwargs: 其他参数

            - symbols: 优化标的列表
            - candidate_events: 候选信号列表
            - results_path: 优化结果保存路径
            - files_position: 优化入场信号文件路径列表
            - signals_module_name: 信号模块名
            - bar_sdt: K线数据开始日期
            - bar_edt: K线数据结束日期

        """
        self.version = 'OpensOptimizeV230924'
        self.symbols = sorted(kwargs['symbols'])
        self.read_bars = read_bars
        self.kwargs = kwargs
        self.task_name = kwargs.get('task_name', '出场优化')
        self.candidate_signals = sorted(kwargs.pop('candidate_signals'))
        self.task_hash = hashlib.md5(f"{self.candidate_signals}_{self.symbols}".encode('utf-8')).hexdigest()[:8].upper()

        results_path = os.path.join(kwargs['results_path'], f"{self.task_name}_{self.task_hash}")
        os.makedirs(results_path, exist_ok=True)
        self.poss_path = os.path.join(results_path, 'poss')
        os.makedirs(self.poss_path, exist_ok=True)

        self.results_path = results_path
        logger.add(f"{self.results_path}\\信号优化.log", encoding='utf-8', enqueue=True)
        logger.info(f"{self.task_name} | {self.candidate_signals} | 其他参数：{kwargs}")

    def _one_symbol_optim(self, symbol):
        one_symbol_optim(symbol, self.read_bars, self.poss_path, optim_type='open',
                         candidate_signals=self.candidate_signals, **self.kwargs)

    def _one_pos_stats(self, pos_name):
        return one_position_stats(self.poss_path, pos_name)

    def __symbols_optim(self, n_jobs=1):
        symbols = self.symbols
        if n_jobs <= 1:
            for symbol in tqdm(sorted(symbols), desc="优化进度"):
                self._one_symbol_optim(symbol)
            return

        with ProcessPoolExecutor(n_jobs) as pool:
            pool.map(self._one_symbol_optim, sorted(symbols))

    def _positions_stats(self, dumps_map, n_jobs=1):
        """统计所有 pos 的表现"""
        if n_jobs <= 1:
            all_stats = [self._one_pos_stats(pos_name) for pos_name in dumps_map.keys()]
            for s, pos_name in zip(all_stats, dumps_map.keys()):
                if not s:
                    continue
                s['pos_dump'] = dumps_map[pos_name]
            return all_stats

        all_stats = []
        with ProcessPoolExecutor(n_jobs) as pool:
            futures = [pool.submit(self._one_pos_stats, pos_name) for pos_name in dumps_map.keys()]
            for future in as_completed(futures):
                s = future.result()
                if s:
                    s['pos_dump'] = dumps_map[s['pos_name']]
                    all_stats.append(s)
        return all_stats

    def execute(self, n_jobs=1):
        """批量优化策略

        :param n_jobs: 进程数量
        :return:
        """
        symbols, results_path = self.symbols, self.results_path
        tactic = CzscOpenOptimStrategy(symbol='symbol', candidate_signals=self.candidate_signals, **self.kwargs)
        tactic.save_positions(os.path.join(results_path, 'positions'))

        dumps_map = {pos.name: pos.dump() for pos in tactic.positions}
        logger.info(f"{self.version} 开始优化策略，策略数量：{len(tactic.positions)}，共 {len(symbols)} 只标的，进程数量：{n_jobs}；"
                    f"结果保存在 {results_path}，请耐心等待...")

        self.__symbols_optim(n_jobs=n_jobs)
        all_stats = self._positions_stats(dumps_map, n_jobs=n_jobs)

        file_report = os.path.join(results_path, f"入场优化_{self.task_name}_{self.task_hash}.xlsx")
        all_stats = [s for s in all_stats if s is not None]
        if all_stats:
            logger.info(f"优化完成，共 {len(all_stats)} 个策略，结果保存在 {file_report}")
            report_df = pd.DataFrame(all_stats).sort_values(['截面等权收益'], ascending=False, ignore_index=True)
            report_df.to_excel(file_report, index=False)
        else:
            logger.warning("优化结果为空！")

        if self.kwargs.get('feishu_app_id') and self.kwargs.get('feishu_app_secret'):
            if os.path.exists(file_report):
                czsc.fsa.push_message(file_report, msg_type='file', **self.kwargs)
            else:
                czsc.fsa.push_message(f"开仓策略优化任务【{self.task_name} 优化结果为空！", **self.kwargs)


class ExitsOptimize:
    """基础策略出场优化流程"""

    def __init__(self, read_bars: Callable, **kwargs):
        """

        :param read_bars: K线数据读取函数
        :param kwargs: 其他参数

            - symbols: 优化标的列表
            - candidate_events: 优化出场信号列表
            - results_path: 优化结果保存路径
            - files_position: 优化入场信号文件路径列表
            - signals_module_name: 信号模块名

        """
        self.version = 'ExitsOptimizeV230924'
        self.symbols = kwargs['symbols']
        self.read_bars = read_bars
        self.kwargs = kwargs
        self.task_name = kwargs.get('task_name', '出场优化')
        self.candidate_events = kwargs.pop('candidate_events')
        self.task_hash = hashlib.md5(f"{self.candidate_events}_{self.symbols}".encode('utf-8')).hexdigest()[:8].upper()

        results_path = os.path.join(kwargs['results_path'], f"{self.task_name}_{self.task_hash}")
        os.makedirs(results_path, exist_ok=True)
        self.poss_path = os.path.join(results_path, 'poss')
        os.makedirs(self.poss_path, exist_ok=True)

        self.results_path = results_path
        logger.add(f"{self.results_path}\\信号优化.log", encoding='utf-8', enqueue=True)
        logger.info(f"{self.task_name} | {self.candidate_events} | 其他参数：{kwargs}")

    def _one_symbol_optim(self, symbol):
        one_symbol_optim(symbol, self.read_bars, self.poss_path, optim_type='exit',
                         candidate_events=self.candidate_events, **self.kwargs)

    def _one_pos_stats(self, pos_name):
        return one_position_stats(self.poss_path, pos_name)

    def __symbols_optim(self, n_jobs=1):
        symbols = self.symbols
        if n_jobs <= 1:
            for symbol in tqdm(sorted(symbols), desc="优化进度"):
                self._one_symbol_optim(symbol)
            return

        with ProcessPoolExecutor(n_jobs) as pool:
            pool.map(self._one_symbol_optim, sorted(symbols))

    def _positions_stats(self, dumps_map, n_jobs=1):
        """统计所有 pos 的表现"""
        if n_jobs <= 1:
            all_stats = [self._one_pos_stats(pos_name) for pos_name in dumps_map.keys()]
            for s, pos_name in zip(all_stats, dumps_map.keys()):
                if not s:
                    continue
                s['pos_dump'] = dumps_map[pos_name]
            return all_stats

        all_stats = []
        with ProcessPoolExecutor(n_jobs) as pool:
            futures = [pool.submit(self._one_pos_stats, pos_name) for pos_name in dumps_map.keys()]
            for future in as_completed(futures):
                s = future.result()
                if s:
                    s['pos_dump'] = dumps_map[s['pos_name']]
                    all_stats.append(s)
        return all_stats

    def execute(self, n_jobs=1):
        """批量优化策略

        :param n_jobs: 进程数量
        :return:
        """
        symbols = self.symbols
        results_path = self.results_path
        tactic = CzscExitOptimStrategy(symbol='symbol', candidate_events=self.candidate_events, **self.kwargs)
        tactic.save_positions(os.path.join(results_path, 'positions'))

        dumps_map = {pos.name: pos.dump() for pos in tactic.positions}
        logger.info(f"{self.version} 开始优化策略，策略数量：{len(tactic.positions)}，共 {len(symbols)} 只标的，进程数量：{n_jobs}；"
                    f"结果保存在 {results_path}，请耐心等待...")

        self.__symbols_optim(n_jobs=n_jobs)
        all_stats = self._positions_stats(dumps_map, n_jobs=n_jobs)

        file_report = os.path.join(results_path, f"出场优化_{self.task_name}_{self.task_hash}.xlsx")
        all_stats = [s for s in all_stats if s is not None]
        if all_stats:
            logger.info(f"策略出场优化完成，共 {len(all_stats)} 个策略，结果保存在 {file_report}")
            report_df = pd.DataFrame(all_stats).sort_values(['截面等权收益'], ascending=False, ignore_index=True)
            report_df.to_excel(file_report, index=False)
        else:
            logger.warning("策略出场优化结果为空！请检查执行日志！")

        if self.kwargs.get('feishu_app_id') and self.kwargs.get('feishu_app_secret'):
            if os.path.exists(file_report):
                czsc.fsa.push_message(file_report, msg_type='file', **self.kwargs)
            else:
                czsc.fsa.push_message("优化结果为空！", **self.kwargs)



def check_missing_values(df): ...
def check_data_types(df): ...
def check_datetime_order(df): ...
def check_price_reasonableness(df): ...
def check_volume_amount(df): ...
def check_symbol_consistency(df): ...
def check_duplicate_records(df): ...
def check_extreme_values(df, threshold: float = 0.2): ...
def check_kline_quality(df): ...



import numpy as np
from collections import Counter as Counter, OrderedDict
from czsc.py.enum import Direction as Direction
from czsc.py.objects import BI as BI, RawBar as RawBar, Signal as Signal, ZS as ZS
from deprecated import deprecated as deprecated
from typing import Any

def create_single_signal(**kwargs) -> OrderedDict: ...
def is_symmetry_zs(bis: list[BI], th: float = 0.3) -> bool: ...
def check_cross_info(fast: list | np.ndarray, slow: list | np.ndarray): ...
def check_gap_info(bars: list[RawBar]): ...
def fast_slow_cross(fast, slow): ...
def same_dir_counts(seq: list | np.ndarray): ...
def count_last_same(seq: list | np.ndarray | tuple): ...
def get_sub_elements(elements: list[Any], di: int = 1, n: int = 10) -> list[Any]: ...
def is_bis_down(bis: list[BI]): ...
def is_bis_up(bis: list[BI]): ...
def get_zs_seq(bis: list[BI]) -> list[ZS]: ...
def cross_zero_axis(n1: list | np.ndarray, n2: list | np.ndarray) -> int: ...
def cal_cross_num(cross: list, distance: int = 1) -> tuple: ...
def down_cross_count(x1: list | np.ndarray, x2: list | np.ndarray) -> int: ...



from _typeshed import Incomplete

def request(method, url, headers, payload=None) -> dict: ...

class FeishuApiBase:
    app_id: Incomplete
    app_secret: Incomplete
    host: str
    headers: Incomplete
    cache: Incomplete
    logger: Incomplete
    def __init__(self, app_id, app_secret, **kwargs) -> None: ...
    def get_access_token(self, key: str = 'app_access_token'): ...
    def get_headers(self): ...
    def get_root_folder_token(self): ...
    def remove(self, token, kind): ...
    def move(self, token, payload): ...
    def copy(self, token, payload): ...
    def upload_file(self, file_path, parent_node): ...
    def download_file(self, file_token, file_path): ...



import pandas as pd

def normalize_feature(df, x_col, method: str = 'standard', **kwargs): ...
def normalize_ts_feature(df, x_col, n: int = 10, **kwargs): ...
def feature_cross_layering(df, x_col, **kwargs): ...
def find_most_similarity(vector: pd.Series, matrix: pd.DataFrame, n: int = 10, metric: str = 'cosine', **kwargs): ...



"""
加密工具模块

提供数据加密和解密功能
"""

from .fernet import (
    generate_fernet_key,
    fernet_encrypt,
    fernet_decrypt,
)

__all__ = [
    'generate_fernet_key',
    'fernet_encrypt',
    'fernet_decrypt',
]



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2022/12/24 22:20
describe: 简单的单仓位策略执行
"""
import os
import webbrowser
import numpy as np
import pandas as pd
from datetime import datetime, timedelta
from collections import OrderedDict
from typing import Callable, List, AnyStr, Union, Optional
from czsc.core import CZSC, Position, RawBar, Signal, BarGenerator
from czsc.utils.data.cache import home_path
from czsc.traders.sig_parse import get_signals_freqs


class CzscSignals:
    """缠中说禅技术分析理论之多级别信号计算"""

    def __init__(self, bg: Optional[BarGenerator] = None, **kwargs):
        """

        :param bg: K线合成器
        """
        self.name = "CzscSignals"
        # cache 是信号计算过程的缓存容器，需要信号计算函数自行维护
        self.cache = OrderedDict()
        self.kwargs = kwargs
        self.signals_config = kwargs.get("signals_config", [])

        if bg:
            self.bg = bg
            assert bg.symbol, "bg.symbol is None"
            self.symbol = bg.symbol
            self.base_freq = bg.base_freq
            self.freqs = list(bg.bars.keys())
            self.kas = {freq: CZSC(b) for freq, b in bg.bars.items()}

            last_bar = self.kas[self.base_freq].bars_raw[-1]
            print(last_bar)
            self.end_dt, self.bid, self.latest_price = last_bar.dt, last_bar.id, last_bar.close
            self.s = OrderedDict()
            self.s.update(self.get_signals_by_conf())
            self.s.update(last_bar.__dict__)
        else:
            self.bg = None
            self.symbol = None
            self.base_freq = None
            self.freqs = None
            self.kas = None
            self.end_dt, self.bid, self.latest_price = None, None, None
            self.s = OrderedDict()

    def __repr__(self):
        return "<{} for {}>".format(self.name, self.symbol)

    def get_signals_by_conf(self):
        """通过信号参数配置获取信号

        函数执行逻辑：

        1. 函数首先创建一个空的有序字典s。
        2. 如果self.signals_config不存在，函数直接返回空字典s，否则，函数遍历其中的每一个配置。
        3. 对于每一个参数，函数提取出信号名称和freq，并根据这两个参数获取相应的信号，获取到的信号被添加到字典s中。
        4. 函数最后返回字典s，其中包含了所有获取到的信号。

        信号参数配置，格式如下：

            signals_config = [
                {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 1, 'ma_type': 'SMA', 'timeperiod': 5},
                {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 5, 'ma_type': 'SMA', 'timeperiod': 5},
                {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 1, 'ma_seq': (5, 20), 'th': 100},
                {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 5, 'ma_seq': (5, 20), 'th': 100},
            ]

        :return: 信号字典
        """
        from czsc.utils import import_by_name

        s = OrderedDict()
        if not self.signals_config:
            return s

        for param in self.signals_config:
            param = dict(param)
            sig_name = param.pop('name')
            sig_func = import_by_name(sig_name) if isinstance(sig_name, str) else sig_name

            freq = param.pop('freq', None)
            if freq in self.kas:    # 如果指定了 freq，那么就使用 CZSC 对象作为输入
                s.update(sig_func(self.kas[freq], **param))
            else:                   # 否则使用 CAT 作为输入
                s.update(sig_func(self, **param))
        return s

    def take_snapshot(self, file_html=None, width: str = "1400px", height: str = "580px"):
        """获取快照

        函数执行逻辑：

        1. 函数首先创建一个Tab对象，用于存储所有的图表和表格。
        2. 函数遍历所有的freq，对于每一个freq，函数获取相应的CZSC对象，并将其转换为一个图表，然后添加到Tab对象中。
        3. 函数提取出所有的信号，并按照freq分组。对于每一个freq，函数创建一个表格，包含该freq下的所有信号，然后添加到Tab对象中。
        4. 如果还有其他的信号，函数创建一个表格，包含所有的其他信号，然后添加到Tab对象中。
        5. 最后，如果提供了file_html参数，函数将Tab对象渲染为一个HTML文件并保存；否则，函数返回Tab对象。

        :param file_html: 交易快照保存的 html 文件名
        :param width: 图表宽度
        :param height: 图表高度
        :return:
        """
        from pyecharts.charts import Tab
        from pyecharts.components import Table
        from pyecharts.options import ComponentTitleOpts

        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            ka: CZSC = self.kas[freq]
            chart = ka.to_echarts(width, height)
            tab.add(chart, freq)

        signals = {k: v for k, v in self.s.items() if len(k.split("_")) == 3}
        for freq in self.freqs:
            # 按各周期K线分别加入信号表
            freq_signals = {k: signals[k] for k in signals.keys() if k.startswith("{}_".format(freq))}
            for k in freq_signals.keys():
                signals.pop(k)
            if len(freq_signals) <= 0:
                continue
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in freq_signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, f"{freq}信号")

        if len(signals) > 0:
            # 加入时间、持仓状态之类的其他信号
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, "其他信号")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def open_in_browser(self, width="1400px", height="580px"):
        """直接在浏览器中打开分析结果

        函数执行逻辑：

        1. 首先创建一个HTML文件的路径file_html，这个文件将被保存在用户的主目录下，文件名为"temp_czsc_advanced_trader.html"。
        2. 然后，函数调用self.take_snapshot方法，将分析结果保存为一个HTML文件。
        3. 最后，函数使用webbrowser.open方法打开这个HTML文件
        """
        file_html = os.path.join(home_path, "temp_czsc_advanced_trader.html")
        self.take_snapshot(file_html, width, height)
        webbrowser.open(file_html)

    def update_signals(self, bar: RawBar):
        """输入基础周期已完成K线，更新信号，更新仓位

        函数执行逻辑：

        1. 函数首先调用self.bg.update(bar)，输入一个已完成的基础周期K线bar，更新各周期K线。
        2. 然后，函数遍历所有的K线freq和对应的K线数据，对每一个K线数据，函数调用self.kas[freq].update(b[-1])，更新对应的 CZSC 对象。
        3. 函数提取出K线的标的代码bar.symbol，并将其赋值给self.symbol。
        4. 函数提取出基础freq的最后一根K线last_bar，并从中提取出结束时间dt，K线IDid，以及收盘价close，并将它们分别赋值给self.end_dt，self.bid，和self.latest_price。
        5. 函数创建一个空的有序字典s，并调用self.get_signals_by_conf()获取所有的信号，然后将这些信号更新到字典s中。
        6. 最后，函数将last_bar的所有属性更新到字典s中。

        :param bar: 基础周期已完成K线
        :return: None
        """
        self.bg.update(bar)
        for freq, b in self.bg.bars.items():
            self.kas[freq].update(b[-1])

        self.symbol = bar.symbol
        last_bar = self.kas[self.base_freq].bars_raw[-1]
        self.end_dt, self.bid, self.latest_price = last_bar.dt, last_bar.id, last_bar.close
        self.s = OrderedDict()
        self.s.update(self.get_signals_by_conf())
        self.s.update(last_bar.__dict__)


def generate_czsc_signals(bars: List[RawBar], signals_config: List[dict],
                          sdt: Union[AnyStr, datetime] = "20170101", init_n: int = 500, df=False, **kwargs):
    """使用 CzscSignals 生成信号

    函数执行逻辑：

    1. 函数首先从信号配置signals_config中获取所有的freqs。
    2. 然后，函数将信号计算开始时间sdt转换为datetime类型，并将开始时间之前的K线数据分配给bars_left，开始时间之后的K线数据分配给bars_right。
    3. 如果bars_right为空，即没有开始时间之后的K线数据，函数会发出一个警告，并返回一个空的DataFrame或空列表。
    4. 函数创建一个BarGenerator对象bg，并使用bars_left中的K线数据来初始化它。
    5. 函数创建一个CzscSignals对象cs，并将bg和信号配置signals_config作为参数传入。
    6. 函数遍历bars_right中的每一根K线，对于每一根K线，函数调用cs.update_signals(bar)来更新信号，并将更新后的信号添加到_sigs列表中。
    7. 最后，如果df参数为True，函数将_sigs转换为DataFrame并返回；否则，直接返回_sigs。

    :param bars: 基础周期 K 线序列
    :param signals_config: 信号函数配置，格式如下：
        signals_config = [
            {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 1, 'ma_type': 'SMA', 'timeperiod': 5},
            {'name': 'czsc.signals.tas_ma_base_V221101', 'freq': '日线', 'di': 5, 'ma_type': 'SMA', 'timeperiod': 5},
            {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 1, 'ma_seq': (5, 20), 'th': 100},
            {'name': 'czsc.signals.tas_double_ma_V221203', 'freq': '日线', 'di': 5, 'ma_seq': (5, 20), 'th': 100},
        ]
    :param sdt: 信号计算开始时间
    :param init_n: 用于 BarGenerator 初始化的基础周期K线数量
    :param df: 是否返回 df 格式的信号计算结果，默认 False
    :return: 信号计算结果
    """
    freqs = get_signals_freqs(signals_config)
    freqs = [freq for freq in freqs if freq != bars[0].freq.value]
    sdt = pd.to_datetime(sdt)                       # type: ignore
    bars_left = [x for x in bars if x.dt < sdt]     # type: ignore
    if len(bars_left) <= init_n:
        bars_left = bars[:init_n]
        bars_right = bars[init_n:]
    else:
        bars_right = [x for x in bars if x.dt >= sdt]   # type: ignore

    if len(bars_right) == 0:
        import warnings
        warnings.warn("右侧K线为空，无法进行信号生成", RuntimeWarning)
        if df:
            return pd.DataFrame()
        else:
            return []

    base_freq = str(bars[0].freq.value)
    bg = BarGenerator(base_freq=base_freq, freqs=freqs, max_count=kwargs.get("bg_max_count", 5000))
    for bar in bars_left:
        bg.update(bar)

    _sigs = []
    cs = CzscSignals(bg, signals_config=signals_config, **kwargs)
    cs.cache.update({'gsc_kwargs': kwargs})
    for bar in bars_right:
        cs.update_signals(bar)
        _sigs.append(dict(cs.s))

    if df:
        return pd.DataFrame(_sigs)
    else:
        return _sigs


def check_signals_acc(bars: List[RawBar], signals_config: List[dict], delta_days: int = 5, **kwargs) -> None:
    """输入基础周期K线和想要验证的信号，输出信号识别结果的快照

    函数执行逻辑：

    1. 函数首先获取基础周期K线的base_freq，并检查输入的K线数据bars是否按时间升序排列。如果bars的长度小于600，函数直接返回。
    2. 然后，函数调用generate_czsc_signals方法，生成Czsc信号，并将结果保存在df中。
    3. 函数提取出df中所有的信号列s_cols，并打印每一列的值的数量。然后，函数将所有的信号添加到signals列表中。
    4. 函数将bars分为两部分，bars_left和bars_right，并获取信号配置signals_config中的所有freqs。
    5. 函数创建一个BarGenerator对象bg，并使用bars_left中的K线数据来初始化它。
    6. 函数创建一个CzscSignals对象ct，并将bg和信号配置signals_config作为参数传入。
    7. 函数创建一个字典last_dt，用于存储每一个信号最后一次出现的时间。
    8. 函数遍历bars_right中的每一根K线，对于每一根K线，函数调用ct.update_signals(bar)来更新信号。
    9. 对于每一个信号，如果当前K线的时间与该信号最后一次出现的时间的差值大于delta_days，并且该信号与当前的信号匹配，
       函数将创建一个HTML文件，保存信号识别结果的快照，并更新该信号最后一次出现的时间。

    :param bars: 原始K线
    :param signals_config: 需要验证的信号列表
    :param delta_days: 两次相同信号之间的间隔天数
    :return: None
    """
    base_freq = str(bars[-1].freq.value)
    assert bars[2].dt > bars[1].dt > bars[0].dt and bars[2].id > bars[1].id, "bars 中的K线元素必须按时间升序"
    if len(bars) < 600:
        return

    df = generate_czsc_signals(bars, signals_config=signals_config, df=True, **kwargs)
    s_cols = [x for x in df.columns if len(x.split("_")) == 3]
    signals = []
    for col in s_cols:
        print('=' * 100, "\n", df[col].value_counts())
        signals.extend([Signal(f"{col}_{v}") for v in df[col].unique() if "其他" not in v])

    print(f"signals: {'+' * 100}")
    for row in signals:
        print(f"- {row}")

    bars_left = bars[:500]
    bars_right = bars[500:]
    freqs = get_signals_freqs(signals_config)
    bg = BarGenerator(base_freq=base_freq, freqs=freqs, max_count=5000)
    for bar in bars_left:
        bg.update(bar)

    ct = CzscSignals(bg, signals_config=signals_config, **kwargs)
    last_dt = {signal.key: ct.end_dt for signal in signals}

    for bar in bars_right:
        ct.update_signals(bar)

        for signal in signals:
            html_path = os.path.join(home_path, signal.key)
            os.makedirs(html_path, exist_ok=True)
            if bar.dt - last_dt[signal.key] > timedelta(days=delta_days) and signal.is_match(ct.s):
                file_html = f"{bar.symbol}_{bar.dt.strftime('%Y%m%d_%H%M')}_{signal.key}_{ct.s[signal.key]}.html"
                file_html = os.path.join(html_path, file_html)
                print(file_html)
                ct.take_snapshot(file_html, height=kwargs.get("height", "680px"))
                last_dt[signal.key] = bar.dt


def get_unique_signals(bars: List[RawBar], signals_config: List[dict], **kwargs):
    """获取信号函数中定义的所有信号列表

    函数执行逻辑：

    1. 函数首先检查输入的K线数据bars是否按时间升序排列。如果bars的长度小于600，函数直接返回一个空列表。
    2. 然后，函数调用generate_czsc_signals方法，生成CZSC信号，并将结果保存在df中。
    3. 函数遍历df中的所有列，对于每一列，如果列名包含三个部分，函数提取出该列中的所有唯一值，然后将列名和每一个唯一值组合成一个新的信号，
        并添加到_res列表中。注意，如果唯一值中包含"其他"，则不会被添加到_res中。
    4. 最后，函数返回_res，其中包含了所有的唯一信号。

    :param bars: 基础K线数据
    :param signals_config: 信号函数配置
    :param kwargs: 传递给generate_czsc_signals方法的参数
    :return: 信号列表
    """
    assert bars[2].dt > bars[1].dt > bars[0].dt and bars[2].id > bars[1].id, "bars 中的K线元素必须按时间升序"
    if len(bars) < 600:
        return []

    df = generate_czsc_signals(bars, signals_config=signals_config, df=True, **kwargs)
    _res = []
    for col in [x for x in df.columns if len(x.split("_")) == 3]:
        _res.extend([f"{col}_{v}" for v in df[col].unique() if "其他" not in v])
    return _res


class CzscTrader(CzscSignals):
    """缠中说禅技术分析理论之多级别联立交易决策类（支持多策略独立执行）"""

    def __init__(self, bg: Optional[BarGenerator] = None, positions: Optional[List[Position]] = None,
                 ensemble_method: Union[AnyStr, Callable] = "mean", **kwargs):
        """

        初始化逻辑：

        1. 首先接收几个参数：
            bg是一个可选的BarGenerator对象，
            positions是一个可选的Position对象列表，
            ensemble_method是一个集成方法，可以是字符串或者一个回调函数。
        2. 函数将positions赋值给self.positions。如果positions不为空，函数会检查positions中的所有名称是否都是唯一的，如果不是，函数会抛出一个断言错误。
        3. 函数将ensemble_method赋值给self.__ensemble_method。这个参数用于指定如何从多个仓位中集成一个仓位。
            它可以是"mean"（平均），"vote"（投票），"max"（最大），或者一个回调函数。
        4. 函数将"CzscTrader"赋值给self.name。
        5. 最后，函数调用父类的初始化函数，传入bg和其他参数。

        :param bg: bar generator 对象
        :param get_signals: 信号计算函数，输入是 CzscSignals 对象，输出是信号字典
        :param ensemble_method: 多个仓位集成一个仓位的方法，可选值 mean, vote, max；也可以传入一个回调函数

            假设有三个仓位对象，当前仓位分别是 1, 1, -1
            mean - 平均仓位，pos = np.mean([1, 1, -1]) = 0.33
            vote - 投票表决，pos = 1
            max  - 取最大，pos = 1

            对于传入回调函数的情况，函数的输入为 dict，key 为 position.name，value 为 position.pos, 样例输入：
            {'多头策略A': 1, '多头策略B': 1, '空头策略A': -1}
        """
        self.positions = positions
        if self.positions:
            _pos_names = [x.name for x in self.positions]
            assert len(_pos_names) == len(set(_pos_names)), "仓位策略名称不能重复"
        self.__ensemble_method = ensemble_method
        self.name = "CzscTrader"
        super().__init__(bg, **kwargs)

    def __repr__(self):
        return "<{} for {}>".format(self.name, self.symbol)

    def update(self, bar: RawBar) -> None:
        """输入基础周期已完成K线，更新信号，更新仓位

        函数执行逻辑：

        1. 函数首先接收一个参数bar，这是一个已完成的基础周期K线。
        2. 函数调用self.update_signals(bar)，输入这个已完成的基础周期K线，更新信号。
        3. 如果self.positions不为空，即存在仓位，函数遍历所有的仓位，对于每一个仓位，函数调用position.update(self.s)，更新该仓位的状态。

        :param bar: 基础周期已完成K线
        :return: None
        """
        self.update_signals(bar)
        if self.positions:
            for position in self.positions:
                position.update(self.s)

    def on_sig(self, sig: dict) -> None:
        """通过信号字典直接交易，用于快速回测场景

        函数执行逻辑：

        1. 函数首先接收一个参数sig，这是一个信号字典，赋值给self.s。
        2. 函数从sig中提取出标的代码symbol，结束时间dt，K线ID id，以及收盘价close，
            并将它们分别赋值给self.symbol，self.end_dt，self.bid，和self.latest_price。
        4. 如果self.positions不为空，即存在持仓策略，函数遍历所有position，函数调用position.update(self.s)，更新该仓位的状态

        :param sig: 信号字典
        :return: None
        """
        self.s = sig
        self.symbol, self.end_dt = self.s['symbol'], self.s['dt']
        self.bid, self.latest_price = self.s['id'], self.s['close']
        if self.positions:
            for position in self.positions:
                position.update(self.s)

    def on_bar(self, bar: RawBar) -> None:
        """输入基础周期已完成K线，更新信号，更新仓位

        :param bar: 基础周期已完成K线
        :return: None
        """
        self.update(bar)

    @property
    def pos_changed(self) -> bool:
        """判断仓位是否发生变化

        1. 函数首先检查self.positions是否为空。如果为空，即没有仓位，函数直接返回False。
        2. 如果self.positions不为空，函数遍历所有的仓位，对于每一个仓位，函数检查其pos_changed属性。
            如果任何一个仓位的pos_changed属性为True，即该仓位发生了变化，函数返回True。

        :return: True/False
        """
        if not self.positions:
            return False
        return any([position.pos_changed for position in self.positions])

    def get_ensemble_pos(self, method: Union[AnyStr, Callable] = None) -> float:
        """获取多个仓位的集成仓位

        函数执行逻辑：

        1. 函数首先检查self.positions是否为空。如果为空，即没有仓位，函数直接返回0。
        2. 如果self.positions不为空，函数获取集成方法method。如果没有传入method参数，函数使用self.__ensemble_method作为集成方法。
        3. 如果method是一个字符串，函数将其转换为小写，然后获取所有仓位的仓位序列pos_seq。
            1. 如果method是"mean"，函数计算pos_seq的平均值作为集成仓位。
            2. 如果method是"vote"，函数计算pos_seq的和的符号作为集成仓位。
            3. 如果method是"max"，函数获取pos_seq的最大值作为集成仓位。
            4. 如果method不是以上任何一个值，函数抛出一个值错误。

        4. 如果method不是一个字符串，即它是一个回调函数，函数将所有仓位的名称和仓位组成的字典作为参数传入method，并将返回值作为集成仓位。

        :param method: 多个仓位集成一个仓位的方法，可选值 mean, vote, max；也可以传入一个回调函数

            假设有三个仓位对象，当前仓位分别是 1, 1, -1
            mean - 平均仓位，pos = np.mean([1, 1, -1]) = 0.33
            vote - 投票表决，pos = 1
            max  - 取最大，pos = 1

            对于传入回调函数的情况，输入是 self.positions

        :return: pos, 集成仓位
        """
        if not self.positions:
            return 0

        method = self.__ensemble_method if not method else method
        if isinstance(method, str):
            method = method.lower()
            pos_seq = [x.pos for x in self.positions]

            if method == 'mean':
                pos = np.mean(pos_seq)
            elif method == 'vote':
                pos = np.sign(sum(pos_seq))
            elif method == 'max':
                pos = max(pos_seq)
            else:
                raise ValueError

        else:
            pos = method({x.name: x.pos for x in self.positions})

        return pos

    def get_position(self, name: str) -> Optional[Position]:
        """获取指定名称的仓位策略对象

        函数执行逻辑：

        1. 函数首先接收一个参数name，这是要查找的仓位名称。
        2. 函数检查self.positions是否为空。如果为空，即没有仓位，函数直接返回None。
        3. 如果self.positions不为空，函数遍历所有的仓位，对于每一个仓位，函数检查其名称是否与输入的名称相同。如果相同，函数返回该仓位。
        4. 如果遍历所有的仓位都没有找到与输入名称相同的仓位，函数返回None。

        :param name: 仓位名称
        :return: Position
        """
        if not self.positions:
            return None

        for position in self.positions:
            if position.name == name:
                return position

        return None

    def take_snapshot(self, file_html=None, width: str = "1400px", height: str = "580px"):
        """获取快照

        :param file_html: 交易快照保存的 html 文件名
        :param width: 图表宽度
        :param height: 图表高度
        :return:
        """
        from pyecharts.charts import Tab
        from pyecharts.components import Table
        from pyecharts.options import ComponentTitleOpts

        tab = Tab(page_title="{}@{}".format(self.symbol, self.end_dt.strftime("%Y-%m-%d %H:%M")))
        for freq in self.freqs:
            ka: CZSC = self.kas[freq]
            bs = None
            if freq == self.base_freq and self.positions:
                # 在基础周期K线上加入最近的操作记录
                bs = []
                for pos in self.positions:
                    for op in pos.operates:
                        if op['dt'] >= ka.bars_raw[0].dt:
                            _op = dict(op)
                            _op['op_desc'] = f"{pos.name} | {_op['op_desc']}"
                            bs.append(_op)

            chart = ka.to_echarts(width, height, bs)
            tab.add(chart, freq)

        signals = {k: v for k, v in self.s.items() if len(k.split("_")) == 3}
        for freq in self.freqs:
            # 按各周期K线分别加入信号表
            freq_signals = {k: signals[k] for k in signals.keys() if k.startswith("{}_".format(freq))}
            for k in freq_signals.keys():
                signals.pop(k)
            if len(freq_signals) <= 0:
                continue
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in freq_signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, f"{freq}信号")

        if len(signals) > 0:
            # 加入时间、持仓状态之类的其他信号
            t1 = Table()
            t1.add(["名称", "数据"], [[k, v] for k, v in signals.items()])
            t1.set_global_opts(title_opts=ComponentTitleOpts(title="缠中说禅信号表", subtitle=""))
            tab.add(t1, "其他信号")

        if file_html:
            tab.render(file_html)
        else:
            return tab

    def get_ensemble_weight(self, method: Optional[Union[AnyStr, Callable]] = None):
        """获取 CzscTrader 中所有 positions 按照 method 方法集成之后的权重

        函数执行逻辑：

        1. 函数首先接收一个参数method，这是集成方法，可以是字符串或者一个回调函数。
        2. 函数检查是否提供了method参数。如果没有提供，函数使用self.__ensemble_method作为集成方法；如果提供了，函数使用提供的method作为集成方法。
        3. 函数调用get_ensemble_weight函数，输入self和method，获取所有仓位按照指定方法集成之后的权重。

        :param method: str or callable
            集成方法，可选值包括：'mean', 'max', 'min', 'vote'
            也可以传入自定义的函数，函数的输入为 dict，key 为 position.name，value 为 position.pos, 样例输入：
                {'多头策略A': 1, '多头策略B': 1, '空头策略A': -1}
        :param kwargs:
        :return: pd.DataFrame
            columns = ['dt', 'symbol', 'weight', 'price']
        """
        from czsc.traders.weight_backtest import get_ensemble_weight
        method = self.__ensemble_method if not method else method
        return get_ensemble_weight(self, method)

    def weight_backtest(self, **kwargs):
        """执行仓位集成权重的回测

        :param kwargs:

            - method: str or callable，集成方法，参考 get_ensemble_weight 方法
            - digits: int，权重小数点后保留的位数，例如 2 表示保留两位小数
            - fee_rate: float，手续费率，例如 0.0002 表示万二

        :return: 回测结果
        """
        from czsc.core import WeightBacktest

        method = kwargs.get("method", self.__ensemble_method)
        digits = kwargs.get("digits", 2)
        fee_rate = kwargs.get("fee_rate", 0.000)
        dfw = self.get_ensemble_weight(method)
        dfw['dt'] = pd.to_datetime(dfw['dt'])
        dfw = dfw[['dt', 'symbol', 'weight', 'price']].copy()
        wb = WeightBacktest(dfw, digits=digits, fee_rate=fee_rate)
        return wb



from __future__ import annotations

from types import ModuleType
from typing import Any

from . import envs as envs
from . import traders as traders
from . import utils as utils
from .core import (
    CZSC,
    Direction,
    Event,
    Freq,
    NewBar,
    Operate,
    Position,
    RawBar,
    Signal,
    ZS,
    format_standard_kline,
)
from .traders import (
    CzscSignals,
    CzscTrader,
    DummyBacktest,
    ExitsOptimize,
    OpensOptimize,
    PairsPerformance,
    SignalsParser,
    check_signals_acc,
    combine_dates_and_pairs,
    combine_holds_and_pairs,
    generate_czsc_signals,
    get_ensemble_weight,
    get_signals_config,
    get_signals_freqs,
    get_unique_signals,
    stoploss_by_direction,
)
from .utils import (
    AliyunOSS,
    DataClient,
    DiskCache,
    clear_cache,
    clear_expired_cache,
    code_namespace,
    create_grid_params,
    cross_sectional_ic,
    dill_dump,
    dill_load,
    disk_cache,
    empty_cache_path,
    fernet_decrypt,
    fernet_encrypt,
    freqs_sorted,
    generate_fernet_key,
    get_dir_size,
    get_py_namespace,
    get_sub_elements,
    get_url_token,
    holds_performance,
    home_path,
    import_by_name,
    index_composition,
    mac_address,
    overlap,
    print_df_sample,
    psi,
    read_json,
    resample_to_daily,
    risk_free_returns,
    rolling_daily_performance,
    save_json,
    set_url_token,
    ta,
    timeout_decorator,
    to_arrow,
    top_drawdowns,
    update_bbars,
    update_nxb,
    update_tbars,
    x_round,
    daily_performance,
)
from .eda import (
    cal_symbols_factor,
    cal_trade_price,
    cal_yearly_days,
    cross_sectional_strategy,
    dif_long_bear,
    judge_factor_direction,
    limit_leverage,
    make_price_features,
    mark_cta_periods,
    mark_volatility,
    min_max_limit,
    monotonicity,
    remove_beta_effects,
    rolling_layers,
    sma_long_bear,
    tsf_type,
    turnover_rate,
    twap,
    unify_weights,
    vwap,
    weights_simple_ensemble,
)
from .features.utils import (
    feature_returns,
    feature_sectional_corr,
    is_event_feature,
    normalize_corr,
)
from .py.bar_generator import check_freq_and_market, get_intraday_times, resample_bars
from .py.calendar import get_trading_dates, is_trading_date, next_trading_date, prev_trading_date
from .sensors import CTAResearch, holds_concepts_effect
from .strategies import CzscJsonStrategy, CzscStrategyBase
from .utils.backtest_report import generate_backtest_report
from .utils.bi_info import calculate_bi_info, symbols_bi_infos
from .utils.kline_quality import check_kline_quality
from .utils.log import log_strategy_info
from .utils.plotting.kline import KlineChart, plot_czsc_chart
from .utils.trade import adjust_holding_weights
from .utils.warning_capture import capture_warnings, execute_with_warning_capture

# 来自可选 Rust 扩展，类型保持宽松
WeightBacktest: Any

__version__: str
__author__: str
__email__: str
__date__: str

# 延迟模块（运行时由 __getattr__ 注入）
svc: ModuleType
fsa: ModuleType
sensors: ModuleType
aphorism: ModuleType
mock: ModuleType
rwc: ModuleType
cwc: ModuleType

__all__ = [
    "envs",
    "traders",
    "utils",
    "CZSC",
    "Direction",
    "Event",
    "Freq",
    "NewBar",
    "Operate",
    "Position",
    "RawBar",
    "Signal",
    "ZS",
    "format_standard_kline",
    "CzscSignals",
    "CzscTrader",
    "DummyBacktest",
    "ExitsOptimize",
    "OpensOptimize",
    "PairsPerformance",
    "SignalsParser",
    "check_signals_acc",
    "combine_dates_and_pairs",
    "combine_holds_and_pairs",
    "generate_czsc_signals",
    "get_ensemble_weight",
    "get_signals_config",
    "get_signals_freqs",
    "get_unique_signals",
    "stoploss_by_direction",
    "AliyunOSS",
    "DataClient",
    "DiskCache",
    "clear_cache",
    "clear_expired_cache",
    "code_namespace",
    "create_grid_params",
    "cross_sectional_ic",
    "dill_dump",
    "dill_load",
    "disk_cache",
    "empty_cache_path",
    "fernet_decrypt",
    "fernet_encrypt",
    "freqs_sorted",
    "generate_fernet_key",
    "get_dir_size",
    "get_py_namespace",
    "get_sub_elements",
    "get_url_token",
    "holds_performance",
    "home_path",
    "import_by_name",
    "index_composition",
    "mac_address",
    "overlap",
    "print_df_sample",
    "psi",
    "read_json",
    "resample_to_daily",
    "risk_free_returns",
    "rolling_daily_performance",
    "save_json",
    "set_url_token",
    "ta",
    "timeout_decorator",
    "to_arrow",
    "top_drawdowns",
    "update_bbars",
    "update_nxb",
    "update_tbars",
    "x_round",
    "daily_performance",
    "WeightBacktest",
    "svc",
    "fsa",
    "sensors",
    "aphorism",
    "mock",
    "rwc",
    "cwc",
    "cal_symbols_factor",
    "cal_trade_price",
    "cal_yearly_days",
    "cross_sectional_strategy",
    "dif_long_bear",
    "judge_factor_direction",
    "limit_leverage",
    "make_price_features",
    "mark_cta_periods",
    "mark_volatility",
    "min_max_limit",
    "monotonicity",
    "remove_beta_effects",
    "rolling_layers",
    "sma_long_bear",
    "tsf_type",
    "turnover_rate",
    "twap",
    "unify_weights",
    "vwap",
    "weights_simple_ensemble",
    "feature_returns",
    "feature_sectional_corr",
    "is_event_feature",
    "normalize_corr",
    "check_freq_and_market",
    "get_intraday_times",
    "resample_bars",
    "get_trading_dates",
    "is_trading_date",
    "next_trading_date",
    "prev_trading_date",
    "CTAResearch",
    "holds_concepts_effect",
    "CzscJsonStrategy",
    "CzscStrategyBase",
    "generate_backtest_report",
    "calculate_bi_info",
    "symbols_bi_infos",
    "check_kline_quality",
    "log_strategy_info",
    "KlineChart",
    "plot_czsc_chart",
    "adjust_holding_weights",
    "capture_warnings",
    "execute_with_warning_capture",
    "welcome",
]


def __getattr__(name: str) -> Any:
    ...


def welcome() -> None:
    ...



import numpy as np
import pandas as pd
from typing import Callable

def vwap(price: np.array, volume: np.array, **kwargs) -> float: ...
def twap(price: np.array, **kwargs) -> float: ...
def remove_beta_effects(df, **kwargs): ...
def cross_sectional_strategy(df, factor, weight: str = 'weight', long: float = 0.3, short: float = 0.3, **kwargs): ...
def judge_factor_direction(df: pd.DataFrame, factor, target: str = 'n1b', by: str = 'symbol', **kwargs): ...
def monotonicity(sequence): ...
def min_max_limit(x, min_val, max_val, digits: int = 4): ...
def rolling_layers(df, factor, n: int = 5, **kwargs): ...
def cal_yearly_days(dts: list, **kwargs): ...
def cal_symbols_factor(dfk: pd.DataFrame, factor_function: Callable, **kwargs): ...
def weights_simple_ensemble(df, weight_cols, method: str = 'mean', only_long: bool = False, **kwargs): ...
def unify_weights(dfw: pd.DataFrame, **kwargs): ...
def sma_long_bear(df: pd.DataFrame, **kwargs): ...
def dif_long_bear(df: pd.DataFrame, **kwargs): ...
def tsf_type(df: pd.DataFrame, factor, n: int = 5, **kwargs): ...
def limit_leverage(df: pd.DataFrame, leverage: float = 1.0, **kwargs): ...
def cal_trade_price(df: pd.DataFrame, digits=None, **kwargs): ...
def mark_cta_periods(df: pd.DataFrame, **kwargs): ...
def mark_v_reversal(df: pd.DataFrame, **kwargs): ...
def mark_volatility(df: pd.DataFrame, kind: str = 'ts', **kwargs): ...
def make_price_features(df, price: str = 'price', **kwargs): ...
def turnover_rate(df: pd.DataFrame, **kwargs): ...



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2023/4/19 23:27
describe: 绩效表现统计
"""
import numpy as np
import pandas as pd
from collections import Counter


def cal_break_even_point(seq) -> float:
    """计算单笔收益序列的盈亏平衡点

    :param seq: 单笔收益序列，数据样例：[0.01, 0.02, -0.01, 0.03, 0.02, -0.02, 0.01, -0.01, 0.02, 0.01]
    :return: 盈亏平衡点
    """
    if sum(seq) < 0:
        return 1.0
    seq = np.cumsum(sorted(seq))  # type: ignore
    return (np.sum(seq < 0) + 1) / len(seq)  # type: ignore


def daily_performance(daily_returns, **kwargs):
    """采用单利计算日收益数据的各项指标

    函数计算逻辑：

    1. 首先，将传入的日收益率数据转换为NumPy数组，并指定数据类型为float64。
    2. 然后，进行一系列判断：如果日收益率数据为空或标准差为零或全部为零，则返回字典，其中所有指标的值都为零。
    3. 如果日收益率数据满足要求，则进行具体的指标计算：

        - 年化收益率 = 日收益率列表的和 / 日收益率列表的长度 * 252
        - 夏普比率 = 日收益率的均值 / 日收益率的标准差 * 标准差的根号252
        - 最大回撤 = 累计日收益率的最高累积值 - 累计日收益率
        - 卡玛比率 = 年化收益率 / 最大回撤（如果最大回撤不为零，则除以最大回撤；否则为10）
        - 日胜率 = 大于零的日收益率的个数 / 日收益率的总个数
        - 年化波动率 = 日收益率的标准差 * 标准差的根号252
        - 下行波动率 = 日收益率中小于零的日收益率的标准差 * 标准差的根号252
        - 非零覆盖 = 非零的日收益率个数 / 日收益率的总个数
        - 回撤风险 = 最大回撤 / 年化波动率；一般认为 1 以下为低风险，1-2 为中风险，2 以上为高风险

    4. 将所有指标的值存储在字典中，其中键为指标名称，值为相应的计算结果。

    :param daily_returns: 日收益率数据，样例：
        [0.01, 0.02, -0.01, 0.03, 0.02, -0.02, 0.01, -0.01, 0.02, 0.01]
    :param kwargs: 其他参数
        - yearly_days: int, 252, 一年的交易日数
    :return: dict
    """
    daily_returns = np.array(daily_returns, dtype=np.float64)
    yearly_days = kwargs.get("yearly_days", 252)

    if len(daily_returns) == 0 or np.std(daily_returns) == 0 or all(x == 0 for x in daily_returns):
        return {
            "绝对收益": 0,
            "年化": 0,
            "夏普": 0,
            "最大回撤": 0,
            "卡玛": 0,
            "日胜率": 0,
            "日盈亏比": 0,
            "日赢面": 0,
            "年化波动率": 0,
            "下行波动率": 0,
            "非零覆盖": 0,
            "盈亏平衡点": 0,
            "新高间隔": 0,
            "新高占比": 0,
            "回撤风险": 0,
        }

    annual_returns = np.sum(daily_returns) / len(daily_returns) * yearly_days
    sharpe_ratio = np.mean(daily_returns) / np.std(daily_returns) * np.sqrt(yearly_days)
    cum_returns = np.cumsum(daily_returns)
    dd = np.maximum.accumulate(cum_returns) - cum_returns
    max_drawdown = np.max(dd)
    kama = annual_returns / max_drawdown if max_drawdown != 0 else 10
    win_pct = len(daily_returns[daily_returns >= 0]) / len(daily_returns)
    daily_mean_loss = np.mean(daily_returns[daily_returns < 0]) if len(daily_returns[daily_returns < 0]) > 0 else 0
    daily_ykb = np.mean(daily_returns[daily_returns >= 0]) / abs(daily_mean_loss) if daily_mean_loss != 0 else 5

    annual_volatility = np.std(daily_returns) * np.sqrt(yearly_days)
    none_zero_cover = len(daily_returns[daily_returns != 0]) / len(daily_returns)

    downside_volatility = np.std(daily_returns[daily_returns < 0]) * np.sqrt(yearly_days)

    # 计算最大新高间隔
    max_interval = Counter(np.maximum.accumulate(cum_returns).tolist()).most_common(1)[0][1]

    # 计算新高时间占比
    high_pct = len([i for i, x in enumerate(dd) if x == 0]) / len(dd)

    def __min_max(x, min_val, max_val, digits=4):
        if x < min_val:
            x1 = min_val
        elif x > max_val:
            x1 = max_val
        else:
            x1 = x
        return round(x1, digits)

    sta = {
        "绝对收益": round(np.sum(daily_returns), 4),
        "年化": round(annual_returns, 4),
        "夏普": __min_max(sharpe_ratio, -5, 10, 2),
        "最大回撤": round(max_drawdown, 4),
        "卡玛": __min_max(kama, -10, 20, 2),
        "日胜率": round(win_pct, 4),
        "日盈亏比": round(daily_ykb, 4),
        "日赢面": round(win_pct * daily_ykb - (1 - win_pct), 4),
        "年化波动率": round(annual_volatility, 4),
        "下行波动率": round(downside_volatility, 4),
        "非零覆盖": round(none_zero_cover, 4),
        "盈亏平衡点": round(cal_break_even_point(daily_returns), 4),
        "新高间隔": max_interval,
        "新高占比": round(high_pct, 4),
        "回撤风险": round(max_drawdown / annual_volatility, 4),
    }
    return sta


def rolling_daily_performance(df: pd.DataFrame, ret_col, window=252, min_periods=100, **kwargs):
    """计算滚动日收益的各项指标

    :param df: pd.DataFrame, 日收益数据，columns=['dt', ret_col] 或者 index 为 datetime64[ns]
    :param ret_col: str, 收益列名
    :param window: int, 滚动窗口, 自然天数
    :param min_periods: int, 最小样本数
    :param kwargs: 其他参数

        - yearly_days: int, 252, 一年的交易日数
    """
    from czsc.eda import cal_yearly_days
    from rs_czsc import daily_performance

    if not df.index.dtype == "datetime64[ns]":
        df["dt"] = pd.to_datetime(df["dt"])
        df.set_index("dt", inplace=True)
    assert df.index.dtype == "datetime64[ns]", "index必须是datetime64[ns]类型, 请先使用 pd.to_datetime 进行转换"

    yearly_days = kwargs.get("yearly_days", cal_yearly_days(df.index.tolist()))

    df = df[[ret_col]].copy().fillna(0)
    df.sort_index(inplace=True, ascending=True)
    dts = sorted(df.index.to_list())
    res = []
    for edt in dts[min_periods:]:
        sdt = edt - pd.Timedelta(days=window)
        dfg = df[(df.index >= sdt) & (df.index <= edt)].copy()
        s = daily_performance(dfg[ret_col].to_list(), yearly_days=yearly_days)
        s["sdt"] = sdt
        s["edt"] = edt
        res.append(s)

    dfr = pd.DataFrame(res)
    return dfr


def evaluate_pairs(pairs: pd.DataFrame, trade_dir: str = "多空") -> dict:
    """评估开平交易记录的表现

    :param pairs: 开平交易记录，数据样例如下：

        ==========  ==========  ===================  ===================  ==========  ==========  ===========  ============  ==========  ==========
        标的代码     交易方向     开仓时间              平仓时间              开仓价格    平仓价格     持仓K线数    事件序列        持仓天数     盈亏比例
        ==========  ==========  ===================  ===================  ==========  ==========  ===========  ============  ==========  ==========
        DLi9001     多头        2019-02-25 21:36:00  2019-02-25 21:51:00     1147.8      1150.72           16  开多 -> 平多           0       25.47
        DLi9001     多头        2021-09-15 14:06:00  2021-09-15 14:09:00     3155.88     3153.61            4  开多 -> 平多           0       -7.22
        DLi9001     多头        2019-08-29 21:01:00  2019-08-29 22:54:00     1445.86     1454.55          114  开多 -> 平多           0       60.09
        DLi9001     多头        2021-10-11 21:46:00  2021-10-11 22:11:00     3631.77     3622.66           26  开多 -> 平多           0      -25.08
        DLi9001     多头        2020-05-13 09:16:00  2020-05-13 09:26:00     1913.13     1917.64           11  开多 -> 平多           0       23.55
        ==========  ==========  ===================  ===================  ==========  ==========  ===========  ============  ==========  ==========

    :param trade_dir: 交易方向，可选值 ['多头', '空头', '多空']
    :return: 交易表现
    """
    assert trade_dir in [
        "多头",
        "空头",
        "多空",
    ], "trade_dir 参数错误，可选值 ['多头', '空头', '多空']"

    pairs = pairs.copy()

    p = {
        "交易方向": trade_dir,
        "交易次数": 0,
        "累计收益": 0,
        "单笔收益": 0,
        "盈利次数": 0,
        "累计盈利": 0,
        "单笔盈利": 0,
        "亏损次数": 0,
        "累计亏损": 0,
        "单笔亏损": 0,
        "交易胜率": 0,
        "累计盈亏比": 0,
        "单笔盈亏比": 0,
        "盈亏平衡点": 1,
        "持仓天数": 0,
        "持仓K线数": 0,
    }

    if len(pairs) == 0:
        return p

    if trade_dir in ["多头", "空头"]:
        pairs = pairs[pairs["交易方向"] == trade_dir]
        if len(pairs) == 0:
            return p

    pairs = pairs.to_dict(orient="records")
    p["交易次数"] = len(pairs)
    p["盈亏平衡点"] = round(cal_break_even_point([x["盈亏比例"] for x in pairs]), 4)
    p["累计收益"] = round(sum([x["盈亏比例"] for x in pairs]), 2)
    p["单笔收益"] = round(p["累计收益"] / p["交易次数"], 2)
    p["持仓天数"] = round(sum([x["持仓天数"] for x in pairs]) / len(pairs), 2)
    p["持仓K线数"] = round(sum([x["持仓K线数"] for x in pairs]) / len(pairs), 2)

    win_ = [x for x in pairs if x["盈亏比例"] >= 0]
    if len(win_) > 0:
        p["盈利次数"] = len(win_)
        p["累计盈利"] = sum([x["盈亏比例"] for x in win_])
        p["单笔盈利"] = round(p["累计盈利"] / p["盈利次数"], 4)
        p["交易胜率"] = round(p["盈利次数"] / p["交易次数"], 4)

    loss_ = [x for x in pairs if x["盈亏比例"] < 0]
    if len(loss_) > 0:
        p["亏损次数"] = len(loss_)
        p["累计亏损"] = sum([x["盈亏比例"] for x in loss_])
        p["单笔亏损"] = round(p["累计亏损"] / p["亏损次数"], 4)

        p["累计盈亏比"] = round(p["累计盈利"] / abs(p["累计亏损"]), 4)
        p["单笔盈亏比"] = round(p["单笔盈利"] / abs(p["单笔亏损"]), 4)

    return p


def holds_performance(df, **kwargs):
    """组合持仓权重表现

    :param df: pd.DataFrame, columns=['dt', 'symbol', 'weight', 'n1b']
        数据说明，dt: 交易时间，symbol: 标的代码，weight: 权重，n1b: 名义收益率
        必须是每个时间点都有数据，如果某个时间点没有数据，可以增加一行数据，权重为0
    :param kwargs:

        - fee: float, 单边费率，BP
        - digits: int, 保留小数位数

    :return: pd.DataFrame, columns=['date', 'change', 'edge_pre_fee', 'cost', 'edge_post_fee']
    """
    fee = kwargs.get("fee", 15)
    digits = kwargs.get("digits", 2)  # 保留小数位数

    df = df.copy()
    df["weight"] = df["weight"].round(digits)
    df = df.sort_values(["dt", "symbol"]).reset_index(drop=True)

    dft = pd.pivot_table(df, index="dt", columns="symbol", values="weight", aggfunc="sum").fillna(0)
    df_turns = dft.diff().abs().sum(axis=1).reset_index()
    df_turns.columns = ["date", "change"]
    sdt = df["dt"].min()
    df_turns.loc[(df_turns["date"] == sdt), "change"] = df[df["dt"] == sdt]["weight"].sum()

    df_edge = df.groupby("dt")[['weight', 'n1b']].apply(lambda x: (x["weight"] * x["n1b"]).sum()).reset_index()
    df_edge.columns = ["date", "edge_pre_fee"]
    dfr = pd.merge(df_turns, df_edge, on="date", how="left")
    dfr["cost"] = dfr["change"] * fee / 10000  # 换手成本
    dfr["edge_post_fee"] = dfr["edge_pre_fee"] - dfr["cost"]  # 净收益
    return dfr


def top_drawdowns(returns: pd.Series, top: int = 10) -> pd.DataFrame:
    """分析最大回撤，返回最大回撤的波峰、波谷、恢复日期、回撤天数、恢复天数

    :param returns: pd.Series, 日收益率序列，index为日期
    :param top: int, optional, 返回最大回撤的数量，默认10
    :return: pd.DataFrame
    """
    returns = returns.copy()
    df_cum = returns.cumsum()
    underwater = df_cum - df_cum.cummax()

    drawdowns = []
    for _ in range(top):
        valley = underwater.idxmin()  # end of the period
        peak = underwater[:valley][underwater[:valley] == 0].index[-1]
        try:
            recovery = underwater[valley:][underwater[valley:] == 0].index[0]
        except IndexError:
            recovery = np.nan  # drawdown not recovered

        # Slice out draw-down period
        if not pd.isnull(recovery):
            underwater.drop(underwater[peak:recovery].index[1:-1], inplace=True)
        else:
            # drawdown has not ended yet
            underwater = underwater.loc[:peak]

        drawdown = df_cum.loc[valley] - df_cum.loc[peak]

        drawdown_days = (valley - peak).days
        recovery_days = (recovery - valley).days if not pd.isnull(recovery) else np.nan
        new_high_days = drawdown_days + recovery_days if not pd.isnull(recovery) else np.nan

        drawdowns.append((peak, valley, recovery, drawdown, drawdown_days, recovery_days, new_high_days))
        if (len(returns) == 0) or (len(underwater) == 0) or (np.min(underwater) == 0):
            break

    df_drawdowns = pd.DataFrame(
        drawdowns, columns=["回撤开始", "回撤结束", "回撤修复", "净值回撤", "回撤天数", "恢复天数", "新高间隔"]
    )
    return df_drawdowns


def psi(df: pd.DataFrame, factor, segment, **kwargs):
    """PSI 群体稳定性指标，反映数据在不同分箱中的分布变化

    PSI = ∑(实际占比 - 基准占比) * ln(实际占比 / 基准占比)

    参考：https://zhuanlan.zhihu.com/p/79682292  风控模型—群体稳定性指标(PSI)深入理解应用

    :param df: 数据, 必须包含 dt 和 col 列
    :param factor: 分组因子
    :param segment: 样本分组
    :param kwargs:
    :return: pd.DataFrame
    """
    dfg = df.groupby([factor, segment], observed=False).size().unstack().fillna(0).apply(lambda x: x / x.sum(), axis=0)
    dfg["总体分布"] = df.groupby(factor).size().values / len(df)
    base_col = "总体分布"

    cols = [x for x in dfg.columns if x != base_col]
    for rate_col in cols:
        dfg[f"{rate_col}_PSI"] = np.where(
            (dfg[base_col] != 0) & (dfg[rate_col] != 0),
            (dfg[rate_col] - dfg[base_col]) * np.log((dfg[rate_col] / dfg[base_col])),
            dfg[rate_col] - dfg[base_col],
        )
    psi_cols = [x for x in dfg.columns if x.endswith("_PSI")]
    dfg["PSI"] = dfg[psi_cols].mean(axis=1)
    dfg.loc["总计"] = dfg.sum(axis=0)
    return dfg



import pandas as pd

def is_event_feature(df, col, **kwargs): ...
def rolling_corr(df, col1, col2, window: int = 300, min_periods: int = 100, **kwargs): ...
def rolling_rank(df: pd.DataFrame, col, window: int = 300, min_periods: int = 100, new_col=None, **kwargs): ...
def rolling_norm(df: pd.DataFrame, col, window: int = 300, min_periods: int = 100, new_col=None, **kwargs): ...
def rolling_qcut(df: pd.DataFrame, col, window: int = 300, min_periods: int = 100, new_col=None, **kwargs): ...
def rolling_compare(df, col1, col2, window: int = 300, min_periods: int = 100, new_col=None, **kwargs) -> None: ...
def rolling_scale(df: pd.DataFrame, col: str, window: int = 300, min_periods: int = 100, new_col=None, **kwargs): ...
def rolling_tanh(df: pd.DataFrame, col: str, window: int = 300, min_periods: int = 100, new_col=None, **kwargs): ...
def rolling_slope(df: pd.DataFrame, col: str, window: int = 300, min_periods: int = 100, new_col=None, **kwargs): ...
def normalize_corr(df: pd.DataFrame, fcol, ycol=None, **kwargs): ...
def feature_returns(df, factor, target: str = 'n1b', **kwargs): ...
def feature_sectional_corr(df, factor, target: str = 'n1b', method: str = 'pearson', **kwargs): ...



from _typeshed import Incomplete
from czsc import Freq as Freq, RawBar as RawBar
from tqsdk import BacktestFinished as BacktestFinished, TqAccount as TqAccount, TqApi, TqBacktest as TqBacktest, TqKq as TqKq, TqSim as TqSim

def format_kline(df, freq=...): ...
def is_trading_end(): ...
def create_symbol_trader(api: TqApi, symbol, **kwargs): ...

symbols: Incomplete
future_name_map: Incomplete

def get_symbols(**kwargs): ...
def get_raw_bars(symbol, freq, sdt, edt, fq: str = '前复权', **kwargs): ...
def get_daily_backup(api: TqApi, **kwargs): ...
def is_trade_time(quote, **kwargs): ...
def adjust_portfolio(api: TqApi, portfolio, account=None, **kwargs): ...



import pandas as pd
from _typeshed import Incomplete
from czsc.fsa.base import FeishuApiBase as FeishuApiBase, request as request

class SpreadSheets(FeishuApiBase):
    def __init__(self, app_id, app_secret, **kwargs) -> None: ...
    def create(self, folder_token, title): ...
    def check(self, token): ...
    def get_sheets(self, token): ...
    def get_sheet_meta(self, token, sheet_id): ...
    def update_values(self, token, data): ...
    def update_styles(self, token, data): ...
    def read_sheet(self, token, sheet_id): ...
    def delete_values(self, token, sheet_id) -> None: ...
    def dimension_range(self, token, data): ...
    def update_sheets(self, token, operates): ...
    def add_permissions_member(self, token, doctype, member_type, member_id, perm) -> None: ...
    def append(self, token, sheet_id, df: pd.DataFrame, batch_size: int = 2000, overwrite: bool = True): ...
    def get_sheet_cols(self, token, sheet_id, n: int = 1): ...
    def read_table(self, token, sheet_id): ...
    def get_condition_formats(self, token, sheet_id): ...
    def set_condition_formats(self, token, data): ...
    def set_styles_batch(self, token, data): ...

class SingleSheet(SpreadSheets):
    token: Incomplete
    sheet_id: Incomplete
    def __init__(self, app_id, app_secret, token, sheet_id) -> None: ...
    def get_meta(self): ...
    def get_cols(self, n: int = 1) -> None: ...
    def single_append(self, df: pd.DataFrame, batch_size: int = 2000, overwrite: bool = False): ...
    def single_read_table(self) -> None: ...
    def single_delete_values(self) -> None: ...



# -*- coding: utf-8 -*-
"""
author: zengbin93
email: zeng_bin8888@163.com
create_dt: 2021/7/16 11:51
"""

import os
import time
import json
import shutil
import hashlib
import inspect
import pandas as pd
from pathlib import Path
from typing import Any, Union, AnyStr


def _get_logger():
    from loguru import logger
    return logger


home_path = Path(os.environ.get("CZSC_HOME", os.path.join(os.path.expanduser("~"), ".czsc")))
home_path.mkdir(parents=True, exist_ok=True)


def get_dir_size(path):
    """获取目录大小，单位：Bytes"""
    total = 0
    with os.scandir(path) as it:
        for entry in it:
            if entry.is_file():
                total += entry.stat().st_size
            elif entry.is_dir():
                total += get_dir_size(entry.path)
    return total


def empty_cache_path():
    shutil.rmtree(home_path)
    os.makedirs(home_path, exist_ok=False)
    print(f"已清空缓存文件夹：{home_path}")


class DiskCache:
    def __init__(self, path=None):
        self.path = home_path / "disk_cache" if path is None else Path(path)
        if self.path.is_file():
            raise Exception("path must be a directory, not a file")

        self.path.mkdir(parents=True, exist_ok=True)

    def __str__(self) -> str:
        return "DiskCache: " + str(self.path)

    def is_found(self, k: str, suffix: str = "pkl", ttl=-1) -> bool:
        """判断缓存文件是否存在

        :param k: 缓存文件名
        :param suffix: 缓存文件后缀，支持 pkl, json, txt, csv, xlsx, feather, parquet
        :param ttl: 缓存文件有效期，单位：秒，-1 表示永久有效
        :return: bool
        """
        file = self.path / f"{k}.{suffix}"
        if not file.exists():
            _get_logger().info(f"缓存文件不存在, {file}")
            return False

        if ttl > 0:
            create_time = file.stat().st_mtime
            if (time.time() - create_time) > ttl:
                _get_logger().info(f"缓存文件已过期, {file}")
                os.remove(file)
                return False

        _get_logger().info(f"缓存文件已找到, {file}")
        return True

    def get(self, k: str, suffix: str = "pkl") -> Any:
        """读取缓存文件

        :param k: 缓存文件名
        :param suffix: 缓存文件后缀，支持 pkl, json, txt, csv, xlsx, feather, parquet
        :return: 缓存文件内容
        """
        file = self.path / f"{k}.{suffix}"
        _get_logger().info(f"正在读取缓存记录，地址：{file}")
        if not file.exists():
            _get_logger().warning(f"文件不存在, {file}")
            return None

        if suffix == "pkl":
            import dill
            res = dill.load(open(file, "rb"))
        elif suffix == "json":
            res = json.load(open(file, "r", encoding="utf-8"))
        elif suffix == "txt":
            res = file.read_text(encoding="utf-8")
        elif suffix == "csv":
            res = pd.read_csv(file, encoding="utf-8")
        elif suffix == "xlsx":
            res = pd.read_excel(file)
        elif suffix == "feather":
            res = pd.read_feather(file)
        elif suffix == "parquet":
            res = pd.read_parquet(file)
        else:
            raise ValueError(f"suffix {suffix} not supported")
        return res

    def set(self, k: str, v: Any, suffix: str = "pkl"):
        """写入缓存文件

        :param k: 缓存文件名
        :param v: 缓存文件内容
        :param suffix: 缓存文件后缀，支持 pkl, json, txt, csv, xlsx, feather, parquet
        """
        file = self.path / f"{k}.{suffix}"
        if file.exists():
            _get_logger().info(f"缓存文件 {file} 将被覆盖")

        if suffix == "pkl":
            import dill
            dill.dump(v, open(file, "wb"))

        elif suffix == "json":
            if not isinstance(v, dict):
                raise ValueError("suffix json only support dict")
            json.dump(v, open(file, "w", encoding="utf-8"), ensure_ascii=False, indent=4)

        elif suffix == "txt":
            if not isinstance(v, str):
                raise ValueError("suffix txt only support str")
            file.write_text(v, encoding="utf-8")

        elif suffix == "csv":
            if not isinstance(v, pd.DataFrame):
                raise ValueError("suffix csv only support pd.DataFrame")
            v.to_csv(file, index=False, encoding="utf-8")

        elif suffix == "xlsx":
            if not isinstance(v, pd.DataFrame):
                raise ValueError("suffix xlsx only support pd.DataFrame")
            v.to_excel(file, index=False)

        elif suffix == "feather":
            if not isinstance(v, pd.DataFrame):
                raise ValueError("suffix feather only support pd.DataFrame")
            v.to_feather(file)

        elif suffix == "parquet":
            if not isinstance(v, pd.DataFrame):
                raise ValueError("suffix parquet only support pd.DataFrame")
            v.to_parquet(file)

        else:
            raise ValueError(f"suffix {suffix} not supported")

        _get_logger().info(f"已写入缓存文件：{file}")

    def remove(self, k: str, suffix: str = "pkl"):
        file = self.path / f"{k}.{suffix}"
        _get_logger().info(f"准备删除缓存文件：{file}")
        Path.unlink(file) if Path.exists(file) else None


def disk_cache(path: Union[AnyStr, Path] = home_path, suffix: str = "pkl", ttl: int = -1):
    """缓存装饰器，支持多种数据格式

    :param path: 缓存文件夹父路径，默认为 home_path，每个函数的缓存文件夹为 path/func_name
    :param suffix: 缓存文件后缀，支持 pkl, json, txt, csv, xlsx, feather, parquet
    :param ttl: 缓存文件有效期，单位：秒
    """

    def decorator(func):
        _c = DiskCache(path=Path(path) / func.__name__)

        def cached_func(*args, **kwargs):
            # 如果函数有 ttl 参数，则使用函数的 ttl 参数
            ttl1 = kwargs.pop("ttl", ttl)

            hash_str = f"{func.__name__}{args}{kwargs}"
            
            try:
                code_str = inspect.getsource(func)
            except OSError:
                code_str = func.__name__
            
            k = hashlib.md5((code_str + hash_str).encode("utf-8")).hexdigest().upper()[:8]
            k = f"{k}_{func.__name__}"

            if _c.is_found(k, suffix=suffix, ttl=ttl1):
                output = _c.get(k, suffix=suffix)
                return output

            else:
                output = func(*args, **kwargs)
                _c.set(k, output, suffix=suffix)
                return output

        return cached_func

    return decorator


def clear_expired_cache(path, max_age: int = 3600 * 24 * 30):
    """清空过期缓存文件夹

    :param path: 缓存文件夹路径
    :param max_age: 最大缓存时间，单位：秒
    """
    _get_logger().info(f"开始清理缓存文件夹：{path}, max_age={max_age} 秒")
    path = Path(path)
    for file in path.glob("*.pkl"):
        # 使用修改时间而不是创建时间，因为在Windows上创建时间不能轻易修改
        file_stat = file.stat()
        if (time.time() - file_stat.st_mtime) > max_age:
            file.unlink()
            _get_logger().info(f"已删除过期缓存文件：{file}, 修改时间：{file_stat.st_mtime}，当前时间：{time.time()}")


def clear_cache(path: Union[AnyStr, Path] = home_path, subs=None, recreate=False):
    """清空缓存文件夹

    :param path: 缓存文件夹路径
    :param subs: 需要清空的子文件夹名称，如果为 None，则清空整个文件夹
    :param recreate: 是否重新创建文件夹, True 时会重新创建文件夹, False 时不会重新创建文件夹
    """
    path = Path(path)
    if subs is None:
        shutil.rmtree(path)
        path.mkdir(parents=True, exist_ok=False)
        _get_logger().info(f"已清空缓存文件夹：{path}")
        return

    for sub in subs:
        fpath = path / sub
        if fpath.exists():
            shutil.rmtree(fpath)
            if recreate:
                fpath.mkdir(parents=True, exist_ok=True)
            _get_logger().info(f"已清空缓存文件夹：{fpath}")



from czsc.utils.data.cache import disk_cache as disk_cache

def generate_symbol_kines(symbol, freq, sdt: str = '20100101', edt: str = '20250101', seed: int = 42): ...
def generate_klines(seed: int = 42): ...
def generate_klines_with_weights(seed: int = 42): ...
def generate_ts_factor(seed: int = 42): ...
def generate_cs_factor(seed: int = 42): ...
def generate_strategy_returns(n_strategies: int = 10, n_days=None, seed: int = 42): ...
def generate_portfolio(seed: int = 42): ...
def set_global_seed(seed: int = 42) -> None: ...
def generate_correlation_data(seed: int = 42): ...
def generate_daily_returns(n_strategies: int = 3, seed: int = 42): ...
def generate_statistics_data(seed: int = 42): ...
def generate_event_data(seed: int = 42): ...



import pandas as pd

def cal_break_even_point(seq) -> float: ...
def daily_performance(daily_returns, **kwargs): ...
def rolling_daily_performance(df: pd.DataFrame, ret_col, window: int = 252, min_periods: int = 100, **kwargs): ...
def evaluate_pairs(pairs: pd.DataFrame, trade_dir: str = '多空') -> dict: ...
def holds_performance(df, **kwargs): ...
def top_drawdowns(returns: pd.Series, top: int = 10) -> pd.DataFrame: ...
def psi(df: pd.DataFrame, factor, segment, **kwargs): ...



from _typeshed import Incomplete
from czsc.py.enum import Direction as Direction, Freq as Freq, Mark as Mark, Operate as Operate
from czsc.utils.analysis.corr import single_linear as single_linear
from dataclasses import dataclass, field
from datetime import datetime
from typing import Callable

@dataclass
class Tick:
    symbol: str
    name: str = ...
    price: float = ...
    vol: float = ...

@dataclass
class RawBar:
    symbol: str
    id: int
    dt: datetime
    freq: Freq
    open: float
    close: float
    high: float
    low: float
    vol: float
    amount: float
    cache: dict = field(default_factory=dict)
    @property
    def upper(self): ...
    @property
    def lower(self): ...
    @property
    def solid(self): ...

@dataclass
class NewBar:
    symbol: str
    id: int
    dt: datetime
    freq: Freq
    open: float
    close: float
    high: float
    low: float
    vol: float
    amount: float
    elements: list = field(default_factory=list)
    cache: dict = field(default_factory=dict)
    @property
    def raw_bars(self): ...

@dataclass
class FX:
    symbol: str
    dt: datetime
    mark: Mark
    high: float
    low: float
    fx: float
    elements: list = field(default_factory=list)
    cache: dict = field(default_factory=dict)
    @property
    def new_bars(self): ...
    @property
    def raw_bars(self): ...
    @property
    def power_str(self): ...
    @property
    def power_volume(self): ...
    @property
    def has_zs(self): ...

@dataclass
class FakeBI:
    symbol: str
    sdt: datetime
    edt: datetime
    direction: Direction
    high: float
    low: float
    power: float
    cache: dict = field(default_factory=dict)

def create_fake_bis(fxs: list[FX]) -> list[FakeBI]: ...

@dataclass
class BI:
    symbol: str
    fx_a: FX
    fx_b: FX
    fxs: list
    direction: Direction
    bars: list[NewBar] = field(default_factory=list)
    cache: dict = field(default_factory=dict)
    sdt = ...
    edt = ...
    def __post_init__(self) -> None: ...
    def get_cache_with_default(self, key, default: Callable): ...
    def get_price_linear(self, price_key: str = 'close'): ...
    @property
    def fake_bis(self): ...
    @property
    def high(self): ...
    @property
    def low(self): ...
    @property
    def power(self): ...
    @property
    def power_price(self): ...
    @property
    def power_volume(self): ...
    @property
    def power_snr(self): ...
    @property
    def change(self): ...
    @property
    def SNR(self): ...
    @property
    def slope(self): ...
    @property
    def acceleration(self): ...
    @property
    def length(self): ...
    @property
    def rsq(self): ...
    @property
    def raw_bars(self): ...
    @property
    def hypotenuse(self): ...
    @property
    def angle(self): ...

@dataclass
class ZS:
    bis: list[BI]
    cache: dict = field(default_factory=dict)
    symbol = ...
    def __post_init__(self) -> None: ...
    @property
    def sdt(self): ...
    @property
    def edt(self): ...
    @property
    def sdir(self): ...
    @property
    def edir(self): ...
    @property
    def zz(self): ...
    @property
    def gg(self): ...
    @property
    def zg(self): ...
    @property
    def dd(self): ...
    @property
    def zd(self): ...
    @property
    def is_valid(self): ...

@dataclass
class Signal:
    signal: str = ...
    score: int = ...
    k1: str = ...
    k2: str = ...
    k3: str = ...
    v1: str = ...
    v2: str = ...
    v3: str = ...
    def __post_init__(self) -> None: ...
    @property
    def key(self) -> str: ...
    @property
    def value(self) -> str: ...
    def is_match(self, s: dict) -> bool: ...

@dataclass
class Event:
    operate: Operate
    signals_all: list[Signal] = field(default_factory=list)
    signals_any: list[Signal] = field(default_factory=list)
    signals_not: list[Signal] = field(default_factory=list)
    name: str = ...
    sha256 = ...
    def __post_init__(self) -> None: ...
    @property
    def unique_signals(self) -> list[str]: ...
    def get_signals_config(self, signals_module: str = 'czsc.signals') -> list[dict]: ...
    def is_match(self, s: dict) -> bool: ...
    def dump(self) -> dict: ...
    @classmethod
    def load(cls, raw: dict): ...

def cal_break_even_point(seq: list[float]) -> float: ...

class Position:
    symbol: Incomplete
    opens: Incomplete
    name: Incomplete
    exits: Incomplete
    events: Incomplete
    interval: Incomplete
    timeout: Incomplete
    stop_loss: Incomplete
    T0: Incomplete
    pos_changed: bool
    operates: Incomplete
    holds: Incomplete
    pos: int
    last_event: Incomplete
    last_lo_dt: Incomplete
    last_so_dt: Incomplete
    end_dt: Incomplete
    def __init__(self, symbol: str, opens: list[Event], exits: list[Event] = [], interval: int = 0, timeout: int = 1000, stop_loss: int = 1000, T0: bool = False, name=None) -> None: ...
    @property
    def unique_signals(self) -> list[str]: ...
    def get_signals_config(self, signals_module: str = 'czsc.signals') -> list[dict]: ...
    def dump(self, with_data: bool = False): ...
    @classmethod
    def load(cls, raw: dict): ...
    @property
    def pairs(self): ...
    def evaluate_holds(self, trade_dir: str = '多空') -> dict: ...
    def evaluate(self, trade_dir: str = '多空') -> dict: ...
    def update(self, s: dict): ...



import pandas as pd

def max_draw_down(n1b: list): ...
def turn_over_rate(df_holds: pd.DataFrame) -> tuple[pd.DataFrame, float]: ...
def holds_concepts_effect(holds: pd.DataFrame, concepts: dict, top_n: int = 20, min_n: int = 3, **kwargs): ...



import pandas as pd
from _typeshed import Incomplete

class PairsPerformance:
    df_pairs: Incomplete
    agg_columns: Incomplete
    def __init__(self, df_pairs: pd.DataFrame) -> None: ...
    @staticmethod
    def get_pairs_statistics(df_pairs: pd.DataFrame): ...
    def agg_statistics(self, col: str): ...
    @property
    def basic_info(self): ...
    def agg_to_excel(self, file_xlsx) -> None: ...

def combine_holds_and_pairs(holds, pairs, results_path): ...
def combine_dates_and_pairs(dates: list, pairs: pd.DataFrame, results_path): ...



# -*- coding: utf-8 -*-
"""
模拟数据生成模块
"""
import numpy as np
import pandas as pd
from czsc.utils.data.cache import disk_cache


def generate_symbol_kines(symbol, freq, sdt="20100101", edt="20250101", seed=42):
    """生成单个品种指定频率的K线数据

    Args:
        symbol: 品种代码，如 'AAPL', '000001.SH' 等
        freq: K线频率，支持 '1分钟', '5分钟', '15分钟', '30分钟', '日线'
        sdt: 开始日期，格式为 'YYYYMMDD'，默认 "20100101"
        edt: 结束日期，格式为 'YYYYMMDD'，默认 "20250101"
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含K线数据的DataFrame，列包括dt、symbol、open、close、high、low、vol、amount
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed + hash(symbol) % 1000)

    # 转换日期格式
    start_date = pd.to_datetime(sdt, format="%Y%m%d")
    end_date = pd.to_datetime(edt, format="%Y%m%d")

    # 根据频率生成时间序列
    if freq == "日线":
        dates = pd.date_range(start=start_date, end=end_date, freq="D")
    elif freq in ["1分钟", "5分钟", "15分钟", "30分钟"]:
        # 先生成日期范围
        trading_days = pd.date_range(start=start_date, end=end_date, freq="D")
        dates = []

        # 获取分钟数
        freq_minutes = int(freq.replace("分钟", ""))

        # A股交易时间段
        morning_start = "09:30"
        morning_end = "11:30"
        afternoon_start = "13:00"
        afternoon_end = "15:00"

        for day in trading_days:
            # 上午交易时间
            morning_times = pd.date_range(
                start=f"{day.strftime('%Y-%m-%d')} {morning_start}",
                end=f"{day.strftime('%Y-%m-%d')} {morning_end}",
                freq=f"{freq_minutes}min",
            )

            # 下午交易时间
            afternoon_times = pd.date_range(
                start=f"{day.strftime('%Y-%m-%d')} {afternoon_start}",
                end=f"{day.strftime('%Y-%m-%d')} {afternoon_end}",
                freq=f"{freq_minutes}min",
            )

            dates.extend(morning_times.tolist())
            dates.extend(afternoon_times.tolist())

        dates = pd.DatetimeIndex(dates)
    else:
        raise ValueError(f"不支持的频率: {freq}。支持的频率: 1分钟, 5分钟, 15分钟, 30分钟, 日线")

    # 定义不同的市场阶段，模拟真实市场的周期性变化
    phases = [
        {"name": "熊市", "trend": -0.0008, "volatility": 0.025, "length": 0.3},
        {"name": "震荡", "trend": 0.0002, "volatility": 0.015, "length": 0.2},
        {"name": "牛市", "trend": 0.0012, "volatility": 0.02, "length": 0.3},
        {"name": "调整", "trend": -0.0005, "volatility": 0.02, "length": 0.2},
    ]

    # 初始价格
    base_price = 100.0

    # 市场阶段控制变量
    total_periods = len(dates)
    phase_idx = 0
    phase_periods = 0
    current_phase = phases[phase_idx]

    data = []

    for i, dt in enumerate(dates):
        # 切换市场阶段
        if phase_periods >= total_periods * current_phase["length"]:
            phase_idx = (phase_idx + 1) % len(phases)
            current_phase = phases[phase_idx]
            phase_periods = 0

        # 当前阶段的趋势和波动
        trend = current_phase["trend"]
        volatility = current_phase["volatility"]

        # 对于分钟级数据，调整趋势和波动率
        if freq != "日线":
            freq_minutes = int(freq.replace("分钟", ""))
            # 按分钟级别调整趋势和波动，使日内波动更合理
            trend = trend / (240 / freq_minutes)  # 240是一天的交易分钟数
            volatility = volatility / (240 / freq_minutes) ** 0.5

        # 添加周期性波动，模拟季节性等因素
        if freq == "日线":
            cycle_factor = np.sin(i / 30) * 0.001  # 30天周期
            annual_cycle = np.sin(i / 365) * 0.0005  # 年度周期
        else:
            # 分钟级别的周期性波动
            cycle_factor = np.sin(i / 120) * 0.0005  # 120分钟周期
            annual_cycle = np.sin(i / (365 * 240)) * 0.0002  # 年度周期

        # 随机噪音
        noise = np.random.normal(0, volatility)

        # 计算开盘价和收盘价
        open_price = base_price
        close_price = base_price * (1 + trend + cycle_factor + annual_cycle + noise)

        # 确保价格不会变为负数
        if close_price <= 0:
            close_price = base_price * 0.95

        # 计算日内波动范围，考虑市场波动的合理性
        price_change_ratio = abs(close_price - open_price) / open_price
        if freq == "日线":
            daily_range = base_price * (price_change_ratio + np.random.uniform(0.01, 0.04))
        else:
            # 分钟级别的波动范围更小
            daily_range = base_price * (price_change_ratio + np.random.uniform(0.001, 0.01))

        if close_price > open_price:  # 阳线
            high_price = close_price + daily_range * np.random.uniform(0.1, 0.5)
            low_price = open_price - daily_range * np.random.uniform(0.1, 0.3)
        else:  # 阴线
            high_price = open_price + daily_range * np.random.uniform(0.1, 0.3)
            low_price = close_price - daily_range * np.random.uniform(0.1, 0.5)

        # 确保价格关系正确：high >= max(open, close), low <= min(open, close)
        high_price = max(high_price, open_price, close_price)
        low_price = min(low_price, open_price, close_price)

        # 模拟成交量 - 价格波动大的时候成交量通常也大
        if freq == "日线":
            base_volume = np.random.uniform(100000, 300000)
        else:
            # 分钟级别的成交量更小
            freq_minutes = int(freq.replace("分钟", ""))
            base_volume = np.random.uniform(10000, 50000) * (freq_minutes / 5)  # 基于5分钟调整

        volatility_factor = price_change_ratio * 5  # 波动率影响成交量
        volume_multiplier = 1 + volatility_factor + np.random.uniform(-0.2, 0.2)
        volume = int(base_volume * max(volume_multiplier, 0.3))  # 确保成交量不会过小

        # 计算成交金额（使用平均价格）
        avg_price = (high_price + low_price + open_price + close_price) / 4
        amount = volume * avg_price

        data.append(
            {
                "dt": dt,
                "symbol": symbol,
                "open": round(open_price, 2),
                "close": round(close_price, 2),
                "high": round(high_price, 2),
                "low": round(low_price, 2),
                "vol": volume,
                "amount": round(amount, 2),
            }
        )

        # 更新基准价格为收盘价
        base_price = close_price
        phase_periods += 1

    return pd.DataFrame(data)


@disk_cache(ttl=3600 * 24)
def generate_klines(seed=42):
    """生成K线数据，包含完整的OHLCVA信息（开高低收量额）

    Args:
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含K线数据的DataFrame，列包括dt、symbol、open、close、high、low、vol、amount
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2010-01-01", end="2025-06-08", freq="D")
    symbols = [
        "AAPL",
        "MSFT",
        "GOOGL",
        "AMZN",
        "TSLA",
        "NVDA",
        "META",
        "NFLX",
        "PYPL",
        "INTC",
        "CSCO",
        "IBM",
        "ORCL",
        "SAP",
        "BTC",
        "ETH",
        "000001",
    ]

    # 定义不同的市场阶段，模拟真实市场的周期性变化
    phases = [
        {"name": "熊市", "trend": -0.0008, "volatility": 0.025, "length": 0.3},
        {"name": "震荡", "trend": 0.0002, "volatility": 0.015, "length": 0.2},
        {"name": "牛市", "trend": 0.0012, "volatility": 0.02, "length": 0.3},
        {"name": "调整", "trend": -0.0005, "volatility": 0.02, "length": 0.2},
    ]

    data = []
    for symbol in symbols:
        # 初始价格
        base_price = 100.0

        # 为每个标的设置不同的种子偏移，确保不同标的有不同的走势
        symbol_seed = seed + hash(symbol) % 1000
        np.random.seed(symbol_seed)

        # 市场阶段控制变量
        total_days = len(dates)
        phase_idx = 0
        phase_days = 0
        current_phase = phases[phase_idx]

        for i, dt in enumerate(dates):
            # 切换市场阶段
            if phase_days >= total_days * current_phase["length"]:
                phase_idx = (phase_idx + 1) % len(phases)
                current_phase = phases[phase_idx]
                phase_days = 0

            # 当前阶段的趋势和波动
            trend = current_phase["trend"]
            volatility = current_phase["volatility"]

            # 添加周期性波动，模拟季节性等因素
            cycle_factor = np.sin(i / 30) * 0.001  # 30天周期

            # 添加长期周期，模拟年度周期
            annual_cycle = np.sin(i / 365) * 0.0005  # 年度周期

            # 随机噪音
            noise = np.random.normal(0, volatility)

            # 计算开盘价和收盘价
            open_price = base_price
            close_price = base_price * (1 + trend + cycle_factor + annual_cycle + noise)

            # 确保价格不会变为负数
            if close_price <= 0:
                close_price = base_price * 0.95

            # 计算日内波动范围，考虑市场波动的合理性
            price_change_ratio = abs(close_price - open_price) / open_price
            daily_range = base_price * (price_change_ratio + np.random.uniform(0.01, 0.04))

            if close_price > open_price:  # 阳线
                high_price = close_price + daily_range * np.random.uniform(0.1, 0.5)
                low_price = open_price - daily_range * np.random.uniform(0.1, 0.3)
            else:  # 阴线
                high_price = open_price + daily_range * np.random.uniform(0.1, 0.3)
                low_price = close_price - daily_range * np.random.uniform(0.1, 0.5)

            # 确保价格关系正确：high >= max(open, close), low <= min(open, close)
            high_price = max(high_price, open_price, close_price)
            low_price = min(low_price, open_price, close_price)

            # 模拟成交量 - 价格波动大的时候成交量通常也大
            base_volume = np.random.uniform(100000, 300000)
            volatility_factor = price_change_ratio * 5  # 波动率影响成交量
            volume_multiplier = 1 + volatility_factor + np.random.uniform(-0.2, 0.2)
            volume = int(base_volume * max(volume_multiplier, 0.3))  # 确保成交量不会过小

            # 计算成交金额（使用平均价格）
            avg_price = (high_price + low_price + open_price + close_price) / 4
            amount = volume * avg_price

            data.append(
                {
                    "dt": dt,
                    "symbol": symbol,
                    "open": round(open_price, 2),
                    "close": round(close_price, 2),
                    "high": round(high_price, 2),
                    "low": round(low_price, 2),
                    "vol": volume,
                    "amount": round(amount, 2),
                }
            )

            # 更新基准价格为收盘价
            base_price = close_price
            phase_days += 1

    return pd.DataFrame(data)


def generate_klines_with_weights(seed=42):
    """生成K线数据，包含权重信息"""
    df = generate_klines(seed)
    df["weight"] = np.random.normal(-1, 1, len(df))
    df["weight"] = df["weight"].clip(-1, 1)
    df["price"] = df["close"]
    return df


def generate_ts_factor(seed=42):
    """生成K线数据，包含权重信息"""
    df = generate_klines(seed)
    df["F#SMA#20"] = df.groupby("symbol")["close"].rolling(20).mean().reset_index(drop=True).fillna(0)
    return df


def generate_cs_factor(seed=42):
    """生成截面因子数据"""
    df = generate_klines(seed)
    df["ret20"] = df.groupby("symbol")["close"].pct_change(20).reset_index(drop=True).fillna(0)
    df["F#RPS#20"] = df.groupby("dt")["ret20"].rank(pct=True).reset_index(drop=True).fillna(0)
    df.drop(columns=["ret20"], inplace=True)
    return df


@disk_cache(ttl=3600 * 24)
def generate_strategy_returns(n_strategies=10, n_days=None, seed=42):
    """生成多策略收益数据

    Args:
        n_strategies: 策略数量，默认10个
        n_days: 生成天数，None表示使用全部时间范围
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含策略收益数据的DataFrame，列包括dt、strategy、returns
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2010-01-01", end="2025-06-08", freq="D")
    if n_days and len(dates) > n_days:
        dates = dates[-n_days:]  # 取最近的n_days天
    data = []

    for i in range(n_strategies):
        strategy_name = f"策略_{i+1:02d}"
        # 生成具有不同特征的收益率
        base_return = np.random.normal(0.0005, 0.015, len(dates))
        if i % 3 == 0:  # 每3个策略中有一个表现更好
            base_return += np.random.normal(0.0002, 0.005, len(dates))

        for j, dt in enumerate(dates):
            data.append({"dt": dt, "strategy": strategy_name, "returns": base_return[j]})

    return pd.DataFrame(data)


@disk_cache(ttl=3600 * 24)
def generate_portfolio(seed=42):
    """生成组合数据

    Args:
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含组合和基准收益数据的DataFrame，列包括dt、portfolio、benchmark
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2010-01-01", end="2025-06-08", freq="D")
    portfolio_returns = np.random.normal(0.0008, 0.012, len(dates))
    benchmark_returns = np.random.normal(0.0003, 0.010, len(dates))

    return pd.DataFrame({"dt": dates, "portfolio": portfolio_returns, "benchmark": benchmark_returns})


def set_global_seed(seed=42):
    """设置全局随机数种子

    Args:
        seed: 随机数种子，默认42

    Note:
        调用此函数后，所有使用numpy随机数的函数都会基于这个种子生成数据
        适用于需要统一设置种子的场景
    """
    np.random.seed(seed)


@disk_cache(ttl=3600 * 24)
def generate_correlation_data(seed=42):
    """生成相关性分析数据

    Args:
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含多个具有不同相关性的时间序列
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2020-01-01", end="2023-12-31", freq="D")

    # 创建具有不同相关性的序列
    base_series = np.random.normal(0, 1, len(dates))

    data = pd.DataFrame(
        {
            "dt": dates,
            "series_A": base_series,
            "series_B": 0.7 * base_series + 0.3 * np.random.normal(0, 1, len(dates)),
            "series_C": -0.5 * base_series + 0.5 * np.random.normal(0, 1, len(dates)),
            "series_D": np.random.normal(0, 1, len(dates)),
            "returns_A": np.random.normal(0.0008, 0.015, len(dates)),
            "returns_B": np.random.normal(0.0005, 0.012, len(dates)),
            "returns_C": np.random.normal(0.0003, 0.020, len(dates)),
        }
    )

    return data


@disk_cache(ttl=3600 * 24)
def generate_daily_returns(n_strategies=3, seed=42):
    """生成日收益数据用于收益分析

    Args:
        n_strategies: 策略数量，默认3个
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含多策略日收益的DataFrame，index为日期
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2020-01-01", end="2023-12-31", freq="D")

    data = {}
    for i in range(n_strategies):
        strategy_name = f"strategy_{chr(65+i)}"  # strategy_A, strategy_B, strategy_C
        # 生成具有不同风险收益特征的收益率
        if i == 0:  # 高收益高波动
            returns = np.random.normal(0.0008, 0.015, len(dates))
        elif i == 1:  # 中等收益中等波动
            returns = np.random.normal(0.0005, 0.012, len(dates))
        else:  # 低收益低波动
            returns = np.random.normal(0.0003, 0.010, len(dates))

        data[strategy_name] = returns

    # 添加基准
    data["benchmark"] = np.random.normal(0.0003, 0.010, len(dates))

    return pd.DataFrame(data, index=dates)


@disk_cache(ttl=3600 * 24)
def generate_statistics_data(seed=42):
    """生成统计分析数据

    Args:
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含统计分析所需的多种数据
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2020-01-01", end="2023-12-31", freq="D")

    data = pd.DataFrame(
        {
            "dt": dates,
            "returns": np.random.normal(0.0008, 0.015, len(dates)),
            "factor1": np.random.normal(0, 1, len(dates)),
            "factor2": np.random.normal(0, 1.2, len(dates)),
            "category": np.random.choice(["A", "B", "C"], len(dates)),
            "volume": np.random.randint(1000000, 10000000, len(dates)),
            "price": np.cumsum(np.random.normal(0.1, 2, len(dates))) + 100,
        }
    )

    return data


@disk_cache(ttl=3600 * 24)
def generate_event_data(seed=42):
    """生成事件分析数据

    Args:
        seed: 随机数种子，确保结果可重现，默认42

    Returns:
        pd.DataFrame: 包含事件和特征的DataFrame
    """
    # 设置随机数种子确保结果可重现
    np.random.seed(seed)

    dates = pd.date_range(start="2020-01-01", end="2023-12-31", freq="D")
    symbols = ["AAPL", "MSFT", "GOOGL", "AMZN", "TSLA"]

    data = []
    for symbol in symbols:
        for dt in dates:
            # 事件发生概率为20%
            event_occur = np.random.choice([0, 1], p=[0.8, 0.2])

            data.append(
                {
                    "dt": dt,
                    "symbol": symbol,
                    "event": event_occur,
                    "target": np.random.normal(0.001, 0.02),
                    "feature1": np.random.normal(0, 1),
                    "feature2": np.random.normal(0, 1.5),
                    "feature3": np.random.normal(0, 0.8),
                    "price_change": np.random.normal(0.0005, 0.015),
                }
            )

    return pd.DataFrame(data)



from collections import OrderedDict
from czsc.core import CZSC as CZSC, RawBar as RawBar
from czsc.utils.sig import create_single_signal as create_single_signal, get_sub_elements as get_sub_elements

def update_vol_ma_cache(c: CZSC, ma_type: str, timeperiod: int, **kwargs): ...
def vol_single_ma_V230214(c: CZSC, **kwargs) -> OrderedDict: ...
def vol_double_ma_V230214(c: CZSC, **kwargs) -> OrderedDict: ...
def vol_ti_suo_V221216(c: CZSC, **kwargs) -> OrderedDict: ...
def vol_gao_di_V221218(c: CZSC, **kwargs) -> OrderedDict: ...
def vol_window_V230731(c: CZSC, **kwargs) -> OrderedDict: ...
def vol_window_V230801(c: CZSC, **kwargs) -> OrderedDict: ...



import pandas as pd
from .html_report_builder import HtmlReportBuilder as HtmlReportBuilder
from .pdf_report_builder import PdfReportBuilder as PdfReportBuilder
from .plotting.backtest import get_performance_metrics_cards as get_performance_metrics_cards, plot_backtest_stats as plot_backtest_stats, plot_colored_table as plot_colored_table, plot_long_short_comparison as plot_long_short_comparison
from _typeshed import Incomplete
from typing import Any

def generate_backtest_report(df: pd.DataFrame, output_path: str | None = None, title: str = '权重回测报告', **kwargs) -> str: ...
def generate_html_backtest_report(df: pd.DataFrame, output_path: str | None = None, title: str = '权重回测报告', **kwargs) -> str: ...

class LongShortComparisonChart:
    df: Incomplete
    config: Incomplete
    def __init__(self, df: pd.DataFrame, config: dict[str, Any]) -> None: ...
    def generate(self) -> str: ...

def generate_pdf_backtest_report(df: pd.DataFrame, output_path: str | None = None, title: str = '权重回测报告', **kwargs) -> str: ...
